import numpy as np

from .. import Geometry, Line, LineSegmentMaterial
from ..objects import WorldObject


class BoxHelper(Line):
    """A WorldObject that shows a box-shaped wireframe.

    Commonly used to visualize bounding boxes.

    Parameters
    ----------
    size : float
        The length of the box' edges in local space.
    thickness : float
        The thickness of the lines in (onscreen) pixels.
    color : Color
        The color of the box.

    """

    def __init__(self, size=1.0, thickness=1, color="white"):
        self._size = float(size)

        positions = np.array(
            [
                [0, 0, 0],  # bottom edges
                [1, 0, 0],
                [0, 0, 0],
                [0, 0, 1],
                [1, 0, 1],
                [1, 0, 0],
                [1, 0, 1],
                [0, 0, 1],
                [0, 1, 0],  # top edges
                [1, 1, 0],
                [0, 1, 0],
                [0, 1, 1],
                [1, 1, 1],
                [1, 1, 0],
                [1, 1, 1],
                [0, 1, 1],
                [0, 0, 0],  # side edges
                [0, 1, 0],
                [1, 0, 0],
                [1, 1, 0],
                [0, 0, 1],
                [0, 1, 1],
                [1, 0, 1],
                [1, 1, 1],
            ],
            dtype="f4",
        )
        positions -= 0.5

        positions *= self._size

        geometry = Geometry(positions=positions)
        material = LineSegmentMaterial(thickness=thickness, color=color, aa=True)

        super().__init__(geometry, material)

    def set_transform_by_aabb(self, aabb, scale=1.0):
        """Align with axis aligned bounding box.

        The position and scale attributes will be configured such that the
        helper will match the given bounding box.

        Parameters
        ----------
        aabb : ndarray, [2, 3]
            The bounding box to align with. The two vectors represent the
            minimum and maximum coordinates of the axis-aligned bounding box.
        scale : float
            Scale multiplier of the final wireframe. Useful for adding margin to the
            box.

        """

        aabb = np.asarray(aabb)
        if aabb.shape != (2, 3):
            raise ValueError(
                "The given array does not appear to represent "
                "an axis-aligned bounding box, ensure "
                "the shape is (2, 3). Shape given: "
                f"{aabb.shape}"
            )

        diagonal = aabb[1] - aabb[0]
        center = aabb[0] + diagonal * 0.5
        full_scale = scale * diagonal / self._size

        self.local.position = center
        self.local.scale = full_scale

    def set_transform_by_object(self, object: WorldObject, space="world", scale=1.0):
        """Align with WorldObject.

        Set the position and scale attributes based on the bounding box of
        another object.

        Parameters
        ----------
        object : WorldObject
            The object to wrap inside this wireframe.
        space : str
            If "world", the wire will be aligned to the world's axes. If
            "local", the wire will be aligned to the local axes.
        scale : float
            Scale multiplier of the final wireframe. Useful for adding margin to the
            box.

        Examples
        --------

        World-space bounding box visualization::

            box = gfx.BoxHelper()
            box.set_transform_by_object(mesh)
            scene.add(box)

        Local-space bounding box visualization::

            box = gfx.BoxHelper()
            box.set_transform_by_object(mesh, space="local")
            mesh.add(box)

        """

        aabb = None
        if space not in {"world", "local"}:
            raise ValueError(
                f'Space argument must be either "world"or "local". Given value: {space}'
            )
        if space == "world":
            aabb = object.get_world_bounding_box()
        elif space == "local" and object.geometry is not None:
            aabb = object.get_geometry_bounding_box()
        if aabb is None:
            raise ValueError(
                "No bounding box could be determined "
                "for the given object, it (and its "
                "children) may not define any geometry"
            )
        self.set_transform_by_aabb(aabb, scale)

    def get_world_bounding_box(self):
        return None

    def get_world_bounding_sphere(self):
        return None



import os
import sys
import atexit
import shutil
import tempfile
import importlib.resources


try:
    HOME = os.path.expanduser("~")
except Exception:  # Exceptions thrown by home() are not specified...
    HOME = "/home"  # Just an arbitrary path


def get_resources_dir():
    """Get the path to the directory of builtin resources."""
    if sys.version_info < (3, 9):
        context = importlib.resources.path("pygfx.data_files", "__init__.py")
    else:
        ref = importlib.resources.files("pygfx.data_files") / "__init__.py"
        context = importlib.resources.as_file(ref)
    with context as path:
        pass
    # Return the dir. We assume that the data files are on a normal dir on the fs.
    return str(path.parent)


def get_cache_dir():
    """Get path were we can store our cache data."""
    return _get_data_dir("cache")


def _get_data_dir(xdg_name):
    # Set by user
    dir = os.getenv("PYGFX_DATA_DIR")
    if dir:
        return os.path.abspath(dir)

    # Get user dir
    user_dir = os.path.expanduser("~")
    if not os.path.isdir(user_dir):
        user_dir = "/var/tmp"

    # Get base cache dir
    if sys.platform.startswith("win"):
        roaming = False
        path1, path2 = os.getenv("LOCALAPPDATA"), os.getenv("APPDATA")
        base_dir = (path2 or path1) if roaming else (path1 or path2)
    elif sys.platform.startswith("darwin"):
        base_dir = os.path.join(user_dir, "Library", "Application Support")
    elif sys.platform.startswith(("linux", "freebsd")):
        # https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
        path1 = os.getenv(f"XDG_{xdg_name.upper()}_HOME")
        path2 = os.path.join(HOME, "." + xdg_name.lower())
        base_dir = path1 or path2

    # Fall back to user dir
    if not (base_dir and os.path.isdir(base_dir)):
        base_dir = user_dir

    # Make directory for pygfx
    dir = os.path.join(base_dir, ".pygfx" if base_dir == user_dir else "pygfx")
    try:
        os.makedirs(dir, exist_ok=True)
        if not (os.access(dir, os.W_OK) and os.path.isdir(dir)):
            raise OSError()
    except OSError:
        # If the config or cache directory cannot be created or is not a writable
        # directory, create a temporary one.
        dir = os.environ["PYGFX_DATA_DIR"] = tempfile.mkdtemp(prefix="pygfx-")
        atexit.register(shutil.rmtree, dir)

    return dir



def test_pyi_fonts(pyi_builder):
    pyi_builder.test_source(
        """
        import pygfx as gfx
        ff = gfx.font_manager.select_font("foo", gfx.font_manager.default_font_props)[0][1]
        print(ff.codepoints)
        assert len(ff.codepoints) > 100
        """
    )


def test_pyi_shaders(pyi_builder):
    pyi_builder.test_source(
        """
        from pygfx.renderers.wgpu import load_wgsl
        wgsl = load_wgsl("line.wgsl")
        assert "fn vs_main(" in wgsl
        """
    )



import numpy as np


class Interpolant:
    """Abstract base class of interpolants over parametric samples.

    This code implements an abstract interpolation class Interpolant.
    It provides a generic template method to find the interval in the array of `parameter_positions`
    that the given parameter value t belongs to.

    It adopts the following optimization measures:

        * Utilizing the binary search algorithm to quickly locate the target interval.
        this reduces the time complexity of the lookup from linear to logarithmic.

        * Caching the index of the last lookup `_cached_index`.
        For continuous lookups within the same interval, the cached value can be directly used without the need for a new search.
        And when the lookup point is close to the cached index, the algorithm will scan the adjacent intervals to find the target interval.

        * Handling the boundary cases (when the lookup point is at the beginning or end of the array) with special processing
        avoiding unnecessary binary searches.

    This approach based on binary search and caching can effectively improve the efficiency of interpolation lookups.
    Time complexity is O(1) for linear access crossing at most 2 intervals
    and O(log2(n)) for random access, where n is the number of positions.

    Parameters
    ----------
    parameter_positions : array_like
        One dimensional array, typically the time or a path along a curve defined by the data.
    sample_values : array_like
        The sample values, can have any dimensionality.
    """

    def __init__(self, parameter_positions, sample_values):
        self.parameter_positions = parameter_positions
        self._cached_index = 0
        self.sample_values = sample_values

    def evaluate(self, t):
        pp = self.parameter_positions
        i1 = self._cached_index
        len_pp = len(pp)

        if len_pp == 0:
            raise ValueError("The parameter_positions array is empty.")

        if len_pp == 1:
            return self.sample_values[0]

        t1 = pp[i1] if i1 < len_pp else None
        t0 = pp[i1 - 1] if i1 > 0 else None

        # check if t is in the interval of the cached index, scan the 2 adjacent intervals at most
        if (t1 is not None and t >= t1) or (t1 is None and t >= t0):
            # scan the right side of the interval, at most 2 intervals
            for _ in range(2):
                i1 += 1
                if i1 >= len_pp:
                    # after the end
                    i1 = len_pp
                    self._cached_index = i1
                    return self.sample_values[i1 - 1]

                t0 = t1
                t1 = pp[i1]

                if t < t1:
                    # we have arrived at the sought interval
                    self._cached_index = i1
                    return self._interpolate(i1, t0, t, t1)

            # prepare binary search on the right side of the index
            right = len_pp
        elif t0 is None or t < t0:
            global_t1 = pp[1]  # the first keyframe
            if t < global_t1:
                # when the animation is looped, scan the biggining after the end
                i1 = 2
                t0 = global_t1

            # scan the left side of the interval, at most 2 intervals
            for _ in range(2):
                i1 -= 1
                if i1 <= 0:
                    # before the start
                    self._cached_index = 0
                    return self.sample_values[0]

                # move to left interval
                t1 = t0
                t0 = pp[i1 - 1]
                if t >= t0:
                    # we have arrived at the sought interval
                    self._cached_index = i1
                    return self._interpolate(i1, t0, t, t1)

            # prepare binary search on the left side of the index
            right = i1
            i1 = 0
        else:
            # the interval is cached, just interpolate
            self._cached_index = i1
            return self._interpolate(i1, t0, t, t1)

        # binary search
        while i1 < right:
            mid = (i1 + right) >> 1
            if t < pp[mid]:
                right = mid
            else:
                i1 = mid + 1

        # check boundary cases, again
        if i1 <= 0:
            self._cached_index = 0
            return self.sample_values[0]

        if i1 >= len_pp:
            i1 = len_pp
            self._cached_index = i1
            return self.sample_values[i1 - 1]

        t1 = pp[i1]
        t0 = pp[i1 - 1]

        self._cached_index = i1
        return self._interpolate(i1, t0, t, t1)

    def _interpolate(self, i1, t0, t, t1):
        raise NotImplementedError("call to abstract method")

    def __call__(self, t):
        return self.evaluate(t)


class LinearInterpolant(Interpolant):
    def __init__(self, parameter_positions, sample_values):
        super().__init__(parameter_positions, sample_values)

    def _interpolate(self, i1, t0, t, t1):
        values = self.sample_values

        alpha = (t - t0) / (t1 - t0)

        return values[i1 - 1] * (1 - alpha) + values[i1] * alpha


class QuaternionLinearInterpolant(Interpolant):
    """
    See: https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#interpolation-slerp

    """

    def __init__(self, parameter_positions, sample_values):
        super().__init__(parameter_positions, sample_values)

    def _interpolate(self, i1, t0, t, t1):
        values = self.sample_values

        vk0 = values[i1 - 1]
        vk1 = values[i1]
        alpha = (t - t0) / (t1 - t0)

        dot = np.dot(vk0, vk1)

        if dot < 0:
            vk1 = -vk1
            dot = -dot

        if (
            dot > 0.99
        ):  # When theta is close to zero, spherical linear interpolation turns into regular linear interpolation.
            q = (1 - alpha) * vk0 + alpha * vk1
            return q / np.linalg.norm(q)  # remember to normalize the quaternion

        theta = np.arccos(dot)
        sin_theta = np.sin(theta)

        return (np.sin((1 - alpha) * theta) / sin_theta) * vk0 + (
            np.sin(alpha * theta) / sin_theta
        ) * vk1


class StepInterpolant(Interpolant):
    def __init__(self, parameter_positions, sample_values):
        super().__init__(parameter_positions, sample_values)

    def _interpolate(self, i1, t0, t, t1):
        return self.sample_values[i1 - 1]


class CubicSplineInterpolant(Interpolant):
    """
    See: https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#interpolation-cubic
    Reference: https://github.com/mrdoob/three.js/blob/master/src/math/interpolants/CubicInterpolant.js
    """

    def __init__(self, parameter_positions, sample_values):
        super().__init__(parameter_positions, sample_values)

    def _interpolate(self, i1, t0, t, t1):
        pps = self.parameter_positions
        len_pps = len(pps)

        i_prev = i1 - 2
        i_next = i1 + 1

        if i_prev < 0:
            i_prev = i1
            t_prev = t1
        else:
            t_prev = pps[i_prev]

        if i_next >= len_pps:
            i_next = i1 - 1
            t_next = t0
        else:
            t_next = pps[i_next]

        half_dt = (t1 - t0) * 0.5

        w_p = half_dt / (t0 - t_prev)
        w_n = half_dt / (t_next - t1)

        p = (t - t0) / (t1 - t0)
        pp = p * p
        ppp = pp * p

        s_p = -w_p * ppp + 2 * w_p * pp - w_p * p
        s_0 = (1 + w_p) * ppp + (-1.5 - 2 * w_p) * pp + (-0.5 + w_p) * p + 1
        s_1 = (-1 - w_n) * ppp + (1.5 + w_n) * pp + 0.5 * p
        s_n = w_n * ppp - w_n * pp

        values = self.sample_values

        vk_prev = values[i_prev]
        vk_0 = values[i1 - 1]
        vk_1 = values[i1]
        vk_next = values[i_next]

        return s_p * vk_prev + s_0 * vk_0 + s_1 * vk_1 + s_n * vk_next



"""
This directory contains wgsl files for the shaders. They can be loaded with ``load_wgsl()``.
"""

import sys
import functools
import importlib.resources


@functools.lru_cache(maxsize=None)
def load_wgsl(shader_name, package_name="pygfx.renderers.wgpu.wgsl"):
    """Load wgsl code from pygfx builtin shader snippets."""

    if sys.version_info < (3, 9):
        context = importlib.resources.path(package_name, shader_name)
    else:
        ref = importlib.resources.files(package_name) / shader_name
        context = importlib.resources.as_file(ref)
    with context as path:
        with open(path, "rb") as f:
            return f.read().decode()



"""
Quickly visualize world objects with as
little boilerplate as possible.
"""

import sys
import numpy as np

from ..objects import (
    Background,
    WorldObject,
    Scene,
    AmbientLight,
    DirectionalLight,
    Light,
)

from ..utils import logger
from ..cameras import Camera, PerspectiveCamera
from ..controllers import OrbitController
from ..materials import BackgroundMaterial
from ..renderers import WgpuRenderer
from ..helpers import Stats


class Display:
    """A Helper to display an object or scene

    This class provides the basic scaffolding needed to visualize a given
    WorldObject. To do so the class chooses a sensible default for each part of
    the full setup unless the value is explicitly set by the user or exists as a
    child of ``object``. For example, it will create a camera unless you
    explicitly set a value for camera or if there is (at least) one camera in
    the scene.

    Parameters
    ----------
    canvas : RenderCanvas
        The canvas used to display the object. If both ``renderer`` and
        ``canvas`` are set, then the renderer needs to use the set canvas.
    renderer : gfx.Renderer
        The renderer to use while drawing the scene. If both ``renderer``
        and ``canvas`` are set, then the renderer needs to use the set canvas.
    controller : gfx.Controller
        The camera controller to use.
    camera : gfx.Camera
        The camera to use. If not set, Display will use the first camera
        in the scene graph. If there is none, Display will create one.
    before_render : Callable
        A callback that will be executed during each draw call before a new
        render is made.
    after_render : Callable
        A callback that will be executed during each draw call after a new
        render is made.
    draw_function : Callable
        Replaces the draw callback with a custom one. If set, both
        `before_render` and `after_render` will have no effect.
    stats : bool
        Display performance statistics such as FPS and draw time
        in the corner of the screen. Defaults to False.

    """

    def __init__(
        self,
        canvas=None,
        renderer=None,
        controller=None,
        camera=None,
        before_render=None,
        after_render=None,
        draw_function=None,
        stats=False,
    ) -> None:
        self.canvas = canvas
        self.renderer = renderer
        self.controller = controller
        self.camera = camera
        self.scene = None
        self.before_render = before_render
        self.after_render = after_render
        self.draw_function = draw_function or self.default_draw
        self.stats = stats

    def default_draw(self):
        if self.before_render is not None:
            self.before_render()

        flush = True
        if self.stats:
            flush = False
            self.stats.start()

        self.renderer.render(self.scene, self.camera, flush=flush)

        if self.stats:
            self.stats.stop()
            self.stats.render()

        if self.after_render is not None:
            self.after_render()

        self.renderer.request_draw()

    def show(
        self,
        object: WorldObject,
        up=(0, 1, 0),
    ):
        """Display a WorldObject

        This function provides you with the basic scaffolding to visualize a given
        WorldObject in a new window. While it does add scaffolding, it aims to be
        fully customizable so that you can replace each piece as needed.

        Parameters
        ----------
        object : gfx.WorldObject
            The object to show. If it is not a :class:`gfx.Scene <pygfx.Scene>`
            then Display will wrap it into a new scene containing lights and a
            background.
        up : ndarray
            If set, and ``object`` does not contain a controller, set the camera
            controller's up vector to this value.

        Notes
        -----
        If you want to display multiple objects, use :class:`gfx.Group
        <pygfx.Group>` instead of :class:`gfx.Scene <pygfx.Scene>` if you
        want lights and background to be added.

        """

        if self.canvas and self.canvas.is_closed():
            raise RuntimeError(
                "Can not show a closed canvas. Did you repeatedly call `show`?"
            )

        # Process scene

        if isinstance(object, Scene):
            custom_scene = False
            scene = object
        else:
            custom_scene = True
            scene = Scene()
            scene.add(object)

            dark_gray = np.array((169, 167, 168, 255)) / 255
            light_gray = np.array((100, 100, 100, 255)) / 255

            background = Background(None, BackgroundMaterial(light_gray, dark_gray))
            scene.add(background)
            scene.add(AmbientLight())
        self.scene = scene

        if not any(scene.iter(lambda x: isinstance(x, Light))):
            logger.warning(
                "Your scene does not contain any lights. Some objects may not be visible."
            )

        # Process renderer

        if self.renderer is None and self.canvas is None:
            from rendercanvas.auto import RenderCanvas

            self.canvas = RenderCanvas()
            self.renderer = WgpuRenderer(self.canvas)
        elif self.renderer is not None:
            self.canvas = self.renderer.target
        elif self.canvas is not None:
            self.renderer = WgpuRenderer(self.canvas)
        elif self.canvas != self.renderer.target:
            raise ValueError("Display's render target differs from it's canvas.")
        else:
            pass

        # Process stats

        if self.stats is True:
            self.stats = Stats(self.renderer)

        # Process camera

        existing_camera = next(scene.iter(lambda x: isinstance(x, Camera)), None)
        if self.camera:
            pass
        elif existing_camera is not None:
            self.camera = existing_camera
        else:
            self.camera = PerspectiveCamera(70, 4 / 3)
            self.camera.show_object(object, up=up, scale=1.4)
        if custom_scene:
            cam_has_light = next(self.camera.iter(lambda x: isinstance(x, Light)), None)
            if not cam_has_light:
                self.camera.add(DirectionalLight())
            self.scene.add(self.camera)

        # Process controller

        if self.controller is None:
            self.controller = OrbitController(register_events=self.renderer)
        if not self.controller.cameras:
            self.controller.add_camera(self.camera)

        self.canvas.request_draw(self.draw_function)
        sys.modules[self.canvas.__module__].loop.run()


def show(
    object: WorldObject,
    up=None,
    *,
    canvas=None,
    renderer=None,
    controller=None,
    camera=None,
    before_render=None,
    after_render=None,
    draw_function=None,
):
    """Display a WorldObject

    This function provides you with the basic scaffolding to visualize a given
    WorldObject in a new window. While it does add scaffolding, it aims to be
    fully customizable so that you can replace each piece as needed.

    Parameters
    ----------
    object : gfx.WorldObject
        The object to show. If it is not a :class:`gfx.Scene <pygfx.Scene>`
        then Display will wrap it into a new scene containing lights and a
        background.
    up : ndarray
        If set, and ``object`` does not contain a controller, set the camera
        controller's up vector to this value.
    canvas : RenderCanvas
        The canvas used to display the object. If both ``renderer`` and
        ``canvas`` are set, then the renderer needs to use the set canvas.
    renderer : gfx.Renderer
        The renderer to use while drawing the scene. If both ``renderer``
        and ``canvas`` are set, then the renderer needs to use the set canvas.
    controller : gfx.Controller
        The camera controller to use.
    camera : gfx.Camera
        The camera to use. If not set, Display will use the first camera
        in the scene graph. If there is none, Display will create one.
    before_render : Callable
        A callback that will be executed during each draw call before a new
        render is made.
    after_render : Callable
        A callback that will be executed during each draw call after a new
        render is made.
    draw_function : Callable
        Replaces the draw callback with a custom one. If set, both
        `before_render` and `after_render` will have no effect.

    Notes
    -----
    If you want to display multiple objects, use :class:`gfx.Group
    <pygfx.Group>` instead of :class:`gfx.Scene <pygfx.Scene>` if you
    want lights and background to be added.

    """

    Display(
        canvas, renderer, controller, camera, before_render, after_render, draw_function
    ).show(object, up)



"""
The enums used in pygfx. The enums are all available from the root ``pygfx`` namespace.

.. currentmodule:: pygfx.utils.enums

.. autosummary::
    :toctree: utils/enums
    :template: ../_templates/custom_layout.rst

    BindMode
    ColorMode
    CoordSpace
    EdgeMode
    ElementFormat
    MarkerInt
    MarkerMode
    MarkerShape
    SizeMode
    TextAlign
    TextAnchor
    VisibleSide
    PixelFilter

"""

from wgpu.utils import BaseEnum
from typing import TypeAlias, Literal


__all__ = [
    "BindMode",
    "ColorMode",
    "CoordSpace",
    "EdgeMode",
    "ElementFormat",
    "MarkerInt",
    "MarkerMode",
    "MarkerShape",
    "PixelFilter",
    "SizeMode",
    "TextAlign",
    "TextAnchor",
    "VisibleSide",
]


class Enum(BaseEnum):
    """Enum base class for pygfx."""


class AlphaMethod(Enum):
    """Enum that defines the different alpha methods."""

    opaque = None  #: opaque object
    stochastic = None  #: stochastic transparency
    blended = None  #: per-fragment blending
    weighted = None  #: weighted blending


class AlphaMode(Enum):
    """Emum that defines the predefined modes for for how the alpha value of an object's fragment is used to combine it with the output texture."""

    auto = (
        None  #: use classic blending, while depth_write defaults to True if opacity==1.
    )
    solid = None  #: alpha is ignored.
    solid_premul = None  #: the alpha is multiplied with the color (making it darker).
    dither = None  #: stochastic transparency with blue noise.
    bayer = None  #: stochastic transparency with a Bayer pattern.
    blend = None  #: use classic alpha blending using the over-operator.
    add = None  #: use additive blending that adds the fragment color, multiplied by alpha.
    subtract = None  #: use subtractive blending that removes the fragment color.
    multiply = None  #: use multiplicative blending that multiplies the fragment color.
    weighted_blend = None  #: weighted blended order independent transparency.
    weighted_solid = None  #: fragments are combined based on alpha, but the final alpha is always 1. Great for e.g. image stitching.
    custom = None  #: value to indicate a custom alpha config.


class EdgeMode(Enum):
    centered = None  #: Centered edges (half the width on each side).
    inner = None  #: Inner edges (the width is added to the inside).
    outer = None  #: Outer edges (the width is added to the outside).


class ColorMode(Enum):
    """The ColorMode enum specifies how an object's color is established."""

    auto = None  #: Use (multiply) all the color sources in (material.color, geometry.colors, material.map) if available.
    uniform = None  #: Use the uniform color (usually ``material.color``).
    vertex = None  #: Use the per-vertex color specified in the geometry  (usually  ``geometry.colors``).
    face = None  #: Use the per-face color specified in the geometry  (usually  ``geometry.colors``).
    vertex_map = None  #: Use per-vertex texture coords (``geometry.texcoords``), and sample these in ``material.map``.
    face_map = None  #: Use per-face texture coords (``geometry.texcoords``), and sample these in ``material.map``.
    debug = (
        None  #: Use colors most suitable for debugging. Defined on a per shader basis.
    )


class MarkerMode(Enum):
    """The MarkerMode enum specifies how an object's marker is established."""

    uniform = None  #: Use a uniform marker, specified on the material.
    vertex = None  #: Use a per-vertex marker specified with ``geometry.markers``.


class SizeMode(Enum):
    """The SizeMode enum specifies how an object's size/width/thickness is established."""

    uniform = None  #: Use a uniform size.
    vertex = None  #: Use a per-vertex size specified with ``geometry.sizes``.


class RotationMode(Enum):
    """The RotationMode enum specifies how an object's rotation is established.

    Currently only used for PointsMaterial.
    """

    uniform = None  #: Use a uniform rotation.
    vertex = None  #: Use a per-vertex rotation specified with ``geometry.rotations``.
    curve = None  #: The rotation follows the curve of the line defined by the points (in screen space).


class CoordSpace(Enum):
    """The CoordSpace enum specifies a coordinate space."""

    model = None  #: The space relative to the object. When the object (or a parent) is e.g. scaled with ``wobject.local.scale = 2`` the thing becomes bigger.
    world = None  #: The space of the scene (the root object). Scaling or rotating of objects does not affect the thing's size or orientation.
    screen = None  #: The screen space (in logical pixels). The thing's size is not affected by zooming or scaling.


class MarkerShape(Enum):
    """The MarkerShape enum specifies the shape of a markers in the PointsMarkerMaterial."""

    circle = None  #: ● A circular shape (i.e. a disk).
    ring = None  #: ○ A circular shape with a hole in the middle.
    square = None  #: ■ A big square shape (sized to encompass the circle shape).
    diamond = None  #: ♦ A rotated square (sized to fit inside the circle).
    plus = None  #: + A plus symbol.
    cross = None  #: x A rotated plus symbol.
    asterix = None  #: ✳️ A plus and a cross combined.
    tick = None  #: A tickmark: an infinitely thin line so only the marker edge is drawn. The width and length can be controller with 'edge_width' and 'size' respectively.
    tick_left = None  #: A tickmark that is on the left side of the line (viewed from the line's start).
    tick_right = None  #: A tickmark that is on the right side of the line (viewed from the line's start).
    triangle_up = None  #: ▲
    triangle_down = None  #: ▼
    triangle_left = None  #: ◀
    triangle_right = None  #: ▶
    heart = None  #: ♥
    spade = None  #: ♠
    club = None  #: ♣
    pin = None  #: 📍
    custom = None  #: Custom shape allowing users to provide their own SDF function


class MarkerInt(Enum):
    """The MarkerInt enums maps marker shape names to an integer."""

    circle = 101
    ring = 102
    square = 201
    diamond = 202
    plus = 203
    cross = 204
    asterix = 205
    tick = 206
    tick_left = 207
    tick_right = 208
    triangle_up = 301
    triangle_down = 302
    triangle_left = 303
    triangle_right = 304
    heart = 401
    spade = 402
    club = 403
    pin = 404
    custom = 901


class ElementFormat(Enum):
    """The base elements to specify formats.

    These values can be used to compose formats of various layouts, e.g. 2D
    positions with "2xf4", rgb colors with "3xu8" or matrices with "4x4xf32".
    The purpose is to provide a common representation for simple formats, that
    can be used for buffers, textures and uniform buffers.
    """

    i1 = None  #: A signed 8bit integer.
    u1 = None  #: An unsigned 8-bit integer (byte).
    i2 = None  #: A signed 16-bit integer.
    u2 = None  #: An unsigned 16-bit integer.
    i4 = None  #: A signed 32-bit integer.
    u4 = None  #: An unsigned 32-bit integer.
    f2 = None  #: A 16-bit float.
    f4 = None  #: A 32-bit float.


class VisibleSide(Enum):
    """The VisibleSide enum specifies what side of a mesh is visible.

    Note that this is the inverse of the "CullMode", as it specifies what
    side is visible rather than what side is culled.
    """

    front = None  #: The front is visible.
    back = None  #: The back is visible.
    both = None  #: Both the front and back are visible.


class BindMode(Enum):
    """The BindMode enum specifies how a skinned mesh is bound to its skeleton."""

    attached = (
        "attached"  #: The skinned mesh shares the same world space as the skeleton.
    )
    detached = "detached"  #: The skinned mesh has its own world space.


class TextAlign(Enum):
    """How text is aligned."""

    start = None  #: The same as left if direction is left-to-right and right if direction is right-to-left.
    end = None  #: The same as right if direction is left-to-right and left if direction is right-to-left.
    left = None  #: The text is aligned to the left edge.
    right = None  #: The text is aligned to the right edge.
    center = None  #: The text is centered between the left and right edges.
    justify = None  #: The words are spread to fill the full space set by max_width.
    justify_all = None  #: Like justifym but also justify the last line.
    auto = None  #: Unspecified alignment (use default).


class TextAnchor(Enum):
    """How text is aligned."""

    top_left = "top-left"
    top_center = "top-center"
    top_right = "top-right"

    baseline_left = "baseline-left"
    baseline_center = "baseline-center"
    baseline_right = "baseline-right"

    middle_left = "middle-left"
    middle_center = "middle-center"
    middle_right = "middle-right"

    bottom_left = "bottom-left"
    bottom_center = "bottom-center"
    bottom_right = "bottom-right"


# TODO: I experimented with using a Literal[] here, an idea discussed in https://github.com/pygfx/wgpu-py/issues/720.
# We should eventually use the same approach to all enums (either an Enum class, or Literal type aliases).

PixelFilter: TypeAlias = Literal[
    "nearest", "linear", "tent", "disk", "bspline", "mitchell", "catmull"
]  #:
""" The type of interpolation for flushing the result of a renderer to a target.

The filter is used both when upsampling and downsampling. The recommended (and default) is "mitchell".
Note that when the source and target image are of the same size, the filter is always nearest.

* "nearest": nearest-neighbour interpolation. Note that this introduces aliasing when downsampling.
* "linear": linear interpolation. Note that this introduces aliasing when downsampling.
* "tent": linearly combines samples based on their distance using a smaller kernel than the cubic filters. When upsampling, it does the same a 'linear'.
* "disk": a circular filter shape to show individual pixels in upsampling cases.
* "bspline": cubic spline of the type b-spline, which is rather smooth but has no overshoot.
* "mitchell": cubic spline of the type Mitchel-Netravali, which is optimized for image interpolation.
* "catmull": cubic spline of the type Catmull-Rom, which is sharper, but has more ringing effects.
"""

# NOTE: Don't forget to add new enums to the toctree and __all__



import numpy as np
from .animation_action import AnimationAction
from ..objects import RootEventHandler, Group, Mesh


class AnimationMixer(RootEventHandler):
    """The AnimationMixer serves as a player for animations within a scene, managing the playback state of all animations.

    It handles blending and playing multiple animation actions simultaneously, combining one or more AnimationAction objects
    and controlling transitions and smooth blending between different animations based on time and weight.
    """

    def __init__(self):
        # todo: separate the animation data from model objects, bind them together here in the mixer with a root object when actions are created
        # self._root = root
        super().__init__()
        self._actions = {}
        self._time = 0
        self._time_scale = 1.0

        self.__activated_actions = []

        self.__property_accu_cache = {}  # Cache for the accumulated target value for active actions

        self.__property_ori_cache = {}  # Cache for the original target value

    @property
    def time(self):
        """The global mixer time (in seconds; starting with 0 on the mixer's creation)."""
        return self._time

    @property
    def time_scale(self):
        """The time scale of the mixer. You can use this to control the general speed of all animations.
        Default is 1.0.
        """
        return self._time_scale

    @time_scale.setter
    def time_scale(self, value):
        self._time_scale = value

    def clip_action(self, clip):
        """Create an AnimationAction from an AnimationClip.
        Calling this method several times with the same clip and root parameters always returns the same clip instance.
        """

        if clip in self._actions:
            return self._actions[clip]

        action = AnimationAction(self, clip)
        self._actions[clip] = action

        # cache the original value
        for t in action._clip.tracks:
            key = (t.target, t.path)
            if key not in self.__property_ori_cache:
                ori_value = self._get_path_value(t.target, t.path).copy()
                if isinstance(ori_value, list):
                    ori_value = np.array(ori_value)
                self.__property_ori_cache[key] = ori_value

        return action

    def _activate_action(self, action):
        if not self._is_active_action(action):
            self.__activated_actions.append(action)

    def _deactivate_action(self, action):
        if self._is_active_action(action):
            self.__activated_actions.remove(action)

        # restore the original value
        for t in action._clip.tracks:
            key = (t.target, t.path)
            if key in self.__property_ori_cache:
                self._set_path_value(t.target, t.path, self.__property_ori_cache[key])

    def _is_active_action(self, action):
        return action in self.__activated_actions

    def set_time(self, time):
        """Set the time of the mixer."""
        self._time = 0
        for action in self.__activated_actions:
            action.time = 0

        self.update(time)

    def update(self, dt):
        """Update the mixer with the time delta."""
        self.__property_accu_cache.clear()
        dt = dt * self._time_scale
        self._time += dt
        for action in self.__activated_actions:
            action._update(dt)

        # apply the accumulated value
        components_map = ["position", "rotation", "scale"]
        for target, paths in self.__property_accu_cache.items():
            components = [None, None, None]
            for path, (accu_value, accu_weight) in paths.items():
                if accu_weight < 1:
                    ori_value = self.__property_ori_cache[(target, path)]
                    if path == "rotation":
                        accu_value = self._mix_slerp(
                            accu_value, ori_value, 1 - accu_weight
                        )
                    else:
                        # translation\scale\weights
                        accu_value = self._mix_lerp(
                            accu_value, ori_value, 1 - accu_weight
                        )

                if path in components_map:
                    components[components_map.index(path)] = accu_value
                else:
                    self._set_path_value(target, path, accu_value)

            if any(c is not None for c in components):
                self._set_path_value(target, "components", components)

    def _accumulate(self, target, path, value, weight):
        accu_cache = self.__property_accu_cache
        if target not in accu_cache:
            accu_cache[target] = {}
        if path not in accu_cache[target]:
            accu_cache[target][path] = (value, weight)
        else:
            current_value, current_weight = accu_cache[target][path]
            current_weight += weight
            mix = weight / current_weight

            if path == "rotation":
                current_value = self._mix_slerp(current_value, value, mix)
            else:
                # translation\scale\weights
                current_value = self._mix_lerp(current_value, value, mix)

            accu_cache[target][path] = (current_value, current_weight)

    def _mix_lerp(self, a, b, t):
        return a * (1 - t) + b * t

    def _mix_slerp(self, a, b, t):
        dot = np.dot(a, b)
        if dot < 0:
            a = -a
            dot = -dot
        if dot > 0.99:
            q = self._mix_lerp(a, b, t)
            return q / np.linalg.norm(q)
        theta = np.arccos(dot)
        sin_theta = np.sin(theta)

        return (np.sin((1 - t) * theta) / sin_theta) * a + (
            np.sin(t * theta) / sin_theta
        ) * b

    def _mix_select(self, a, b, t):
        return a if t < 0.5 else b

    def _get_path_value(self, target, path):
        if path == "scale":
            return target.local.scale
        elif path == "translation":
            return target.local.position
        elif path == "rotation":
            return target.local.rotation
        elif path == "components":
            return target.local.components
        elif path == "weights":
            if isinstance(target, Mesh):
                return target.morph_target_influences
            elif isinstance(target, Group):
                for c in target.children:
                    if isinstance(c, Mesh):
                        return c.morph_target_influences

    def _set_path_value(self, target, path, value):
        if path == "scale":
            target.local.scale = value
        elif path == "translation":
            target.local.position = value
        elif path == "rotation":
            target.local.rotation = value
        elif path == "components":
            target.local.components = value
        elif path == "weights":
            if isinstance(target, Mesh):
                target.morph_target_influences = value
            elif isinstance(target, Group):
                for c in target.children:
                    if isinstance(c, Mesh):
                        c.morph_target_influences = value



import jinja2

root_loader = jinja2.PrefixLoader({}, delimiter=".")

jinja_env = jinja2.Environment(
    block_start_string="{$",
    block_end_string="$}",
    variable_start_string="{{",
    variable_end_string="}}",
    line_statement_prefix="$$",
    undefined=jinja2.StrictUndefined,
    loader=root_loader,
)


def register_wgsl_loader(context, loader):
    """Register a source for shader snippets.

    When code is encountered that looks like::

       {$ include 'some_context.name.wgsl' $}

    The loader for "some_context" is looked up and used to load the wgsl to include.
    This function allows registering a loader for your downstream package or application.

    Parameters
    ----------
    context : str
        The context of the loader.
    loader: jinja2.BaseLoader | callable | dict
        The loader to use for this context. If a function is given, it must accept one
        positional argument (the name to include).
    """
    if not (isinstance(context, str) and "." not in context):
        raise TypeError("Wgsl load context must be a string witout dots.")
    if context in root_loader.mapping:
        raise RuntimeError(f"A loader is already registered for '{context}'.")
    if isinstance(loader, jinja2.BaseLoader):
        root_loader.mapping[context] = loader
    elif isinstance(loader, dict):
        root_loader.mapping[context] = jinja2.DictLoader(loader)
    elif callable(loader):
        root_loader.mapping[context] = jinja2.FunctionLoader(loader)
    else:
        raise TypeError(
            f"The given wgsl loader must be a jinja2.BaseLoader, function, or dict. Not {loader!r}"
        )


register_wgsl_loader("pygfx", jinja2.PackageLoader("pygfx.renderers.wgpu.wgsl", "."))


def apply_templating(code, **kwargs):
    t = jinja_env.from_string(code)
    try:
        return t.render(**kwargs)
    except jinja2.UndefinedError as err:
        raise ValueError(f"Cannot compose shader: {err.args[0]}") from None



import wgpu  # only for flags/enums

from ....objects import Image
from ....materials import ImageBasicMaterial
from ....resources import Texture

from .. import (
    register_wgpu_render_function,
    BaseShader,
    Binding,
    to_texture_format,
    GfxSampler,
    GfxTextureView,
    load_wgsl,
)


vertex_and_fragment = wgpu.ShaderStage.VERTEX | wgpu.ShaderStage.FRAGMENT


@register_wgpu_render_function(Image, ImageBasicMaterial)
class ImageShader(BaseShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        material = wobject.material
        geometry = wobject.geometry

        # In this method we want to set all shader variables (i.e. self['xx']),
        # so that we can properly detect when stuff changes that (may) affect
        # these values, and trigger a recompile.

        # Check grid
        if geometry.grid is None:
            raise ValueError("Image.geometry must have a grid (texture).")
        elif not isinstance(geometry.grid, Texture):
            raise TypeError("Image.geometry.grid must be a Texture.")
        elif geometry.grid.dim != 2:
            raise TypeError("Image.geometry.grid must a 2D texture")

        # Set img_format and climcorrection
        self["climcorrection"] = ""
        fmt = to_texture_format(geometry.grid.format)
        if "norm" in fmt or "float" in fmt:
            self["img_format"] = "f32"
            if "unorm" in fmt:
                self["climcorrection"] = " * 255.0"
            elif "snorm" in fmt:
                self["climcorrection"] = " * 255.0 - 128.0"
        elif "uint" in fmt:
            self["img_format"] = "u32"
        else:
            self["img_format"] = "i32"

        # Set gamma
        self["gamma"] = material.gamma

        # Channels
        self["img_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))

        # Determine colorspace
        self["colorspace"] = geometry.grid.colorspace
        self["colorrange"] = geometry.grid.colorrange
        self["three_grid_yuv"] = (
            self["colorspace"] in ["yuv420p", "yuv444p"] and geometry.grid.size[2] == 1
        )
        self["use_colormap"] = False

        if geometry.grid.colorspace not in ("srgb", "physical"):
            # A special color-space that we convert to rgb in the shader
            self["img_nchannels"] = 3
        elif material.map is not None:
            self["use_colormap"] = True
            self["colorspace"] = material.map.texture.colorspace

    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
        ]

        tex_view = GfxTextureView(geometry.grid)
        sampler = GfxSampler(material.interpolation, "clamp")
        bindings.append(Binding("s_img", "sampler/filtering", sampler, "FRAGMENT"))
        bindings.append(Binding("t_img", "texture/auto", tex_view, vertex_and_fragment))

        if self["three_grid_yuv"]:
            u_tex_view = GfxTextureView(geometry.grid_u)
            v_tex_view = GfxTextureView(geometry.grid_v)
            bindings.append(
                Binding("t_u_img", "texture/auto", u_tex_view, vertex_and_fragment)
            )
            bindings.append(
                Binding("t_v_img", "texture/auto", v_tex_view, vertex_and_fragment)
            )

        if self["use_colormap"]:
            bindings.extend(self.define_img_colormap(material.map))

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_strip,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        return {
            "indices": (4, 1),
        }

    def get_code(self):
        return load_wgsl("image.wgsl")



from importlib.util import find_spec

import numpy as np

from ._mesh import MeshStandardMaterial, MeshPhongMaterial
from ..resources import Texture
from ..utils.color import Color


def texture_from_pillow_image(image, dim=2, **kwargs):
    """Pillow Image texture.

    Create a Texture from a PIL.Image.

    Parameters
    ----------
    image : Image
        The `PIL.Image
        <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image>`_
        to convert into a texture.
    dim : int
        The number of spatial dimensions of the image.
    kwargs : Any
        Additional kwargs are forwarded to :class:`pygfx.Texture`.

    Returns
    -------
    image_texture : Texture
        A texture object representing the given image.

    """

    from PIL.Image import Image

    if not isinstance(image, Image):
        raise NotImplementedError()

    # If this is a palette image, convert it to RGBA
    if getattr(image, "mode", None) == "P":
        image = image.convert("RGBA")

    m = memoryview(image.tobytes())

    im_channels = len(image.getbands())
    buffer_shape = (*image.size, im_channels)

    m = m.cast(m.format, shape=buffer_shape)
    return Texture(m, dim=dim, **kwargs)


def material_from_trimesh(x):
    """Convert a Trimesh object into a pygfx material.

    Parameters
    ----------
    x : trimesh.Material | trimesh.Visuals | trimesh.Trimesh
        Either the actual material to convert or an object containing
        the material.

    Returns
    -------
    converted : Material
        The converted material.

    """
    import trimesh
    from trimesh.visual.material import PBRMaterial, SimpleMaterial
    from trimesh.visual import ColorVisuals

    # If this is a trimesh object, extract the visual
    if isinstance(x, trimesh.Trimesh):
        x = x.visual

    # If this is a visual, check if it has a material; this is the case
    # for e.g. TextureVisuals which should contain a SimpleMaterial.
    # ColorVisuals, on the other hand, do not have a material and are
    # typically used to store vertex or face colors - we will deal with
    # them below
    if isinstance(x, trimesh.visual.base.Visuals) and hasattr(x, "material"):
        x = x.material

    if isinstance(x, PBRMaterial):
        material = x
        gfx_material = MeshStandardMaterial()

        if material.baseColorTexture is not None:
            gfx_material.map = texture_from_pillow_image(material.baseColorTexture)

        if material.emissiveFactor is not None:
            gfx_material.emissive = Color(*material.emissiveFactor)

        if material.emissiveTexture is not None:
            gfx_material.emissive_map = texture_from_pillow_image(
                material.emissiveTexture
            )

        if material.metallicRoughnessTexture is not None:
            metallic_roughness_map = texture_from_pillow_image(
                material.metallicRoughnessTexture
            )
            gfx_material.roughness_map = metallic_roughness_map
            gfx_material.metalness_map = metallic_roughness_map
            gfx_material.roughness = 1.0
            gfx_material.metalness = 1.0

        if material.roughnessFactor is not None:
            gfx_material.roughness = material.roughnessFactor

        if material.metallicFactor is not None:
            gfx_material.metalness = material.metallicFactor

        if material.normalTexture is not None:
            gfx_material.normal_map = texture_from_pillow_image(material.normalTexture)
            # See: https://github.com/KhronosGroup/glTF-Sample-Assets/tree/main/Models/NormalTangentTest#problem-flipped-y-axis-or-flipped-green-channel
            gfx_material.normal_scale = (1.0, -1.0)

        if material.occlusionTexture is not None:
            gfx_material.ao_map = texture_from_pillow_image(material.occlusionTexture)

        gfx_material.side = "front"
    elif isinstance(x, SimpleMaterial):
        material = x
        gfx_material = MeshPhongMaterial(color=material.ambient / 255)

        gfx_material.shininess = material.glossiness
        gfx_material.specular = Color(*(material.specular / 255))

        # Note: `material.image` can exist but be None
        if getattr(material, "image", None):
            gfx_material.map = texture_from_pillow_image(material.image)

        gfx_material.side = "front"
    elif isinstance(x, ColorVisuals):
        # ColorVisuals are typically used for vertex or face colors but they can be
        # also be undefined in which case it's just a default material
        gfx_material = MeshPhongMaterial(color=x.main_color / 255)

        default_mat = getattr(trimesh.visual.color, "DEFAULT_MAT", None)
        if default_mat is None:
            default_mat = x.defaults  # trimesh < 4.6.3
        gfx_material.shininess = default_mat["material_shine"]
        gfx_material.specular = Color(*(default_mat["material_specular"] / 255))

        gfx_material.side = "front"

        if x.kind == "vertex":
            gfx_material.color_mode = "vertex"
        elif x.kind == "face":
            gfx_material.color_mode = "face"
    else:
        raise NotImplementedError(f"Conversion of {type(x)} is not supported.")

    return gfx_material


def material_from_open3d(x):
    """Convert an Open3D MaterialRecord object into a pygfx material.

    Parameters
    ----------
    x : open3d.visualization.rendering.MaterialRecord | open3d.geometry.Geometry3D
        Either the actual Open3D MaterialRecord or an object containing
        a material (e.g., geometry).

    Returns
    -------
    converted : Material
        The converted pygfx material.

    """
    if not find_spec("open3d"):
        raise ImportError(
            "The `open3d` library is required for this function: pip install open3d"
        )

    import open3d as o3d

    # Ensure the input is a MaterialRecord
    if not isinstance(x, o3d.visualization.rendering.MaterialRecord):
        raise NotImplementedError("Input must be an Open3D MaterialRecord")

    # Determine which pygfx material to create based on shader type
    if x.shader in ["defaultLit", "litPBR"]:
        gfx_material = MeshStandardMaterial()

        # Set base metallic and roughness values
        gfx_material.metalness = getattr(x, "base_metallic", 1.0)
        gfx_material.roughness = getattr(x, "base_roughness", 1.0)

        # Set albedo color if available
        if x.base_color is not None:
            albedo_color = (
                np.array(x.base_color[:3]) / 255
            )  # Convert to normalized values
            gfx_material.color = Color(*albedo_color)

        # Handle textures if available
        if x.albedo_img is not None:
            gfx_material.map = Texture(np.ascontiguousarray(x.albedo_img), dim=2)

        if x.normal_img is not None:
            gfx_material.normal_map = Texture(np.ascontiguousarray(x.normal_img), dim=2)
            gfx_material.normal_scale = (1.0, -1.0)

        if x.ao_img is not None:
            gfx_material.ao_map = Texture(np.ascontiguousarray(x.ao_img), dim=2)

        if x.metallic_img is not None:
            gfx_material.metalness_map = Texture(
                np.ascontiguousarray(x.metallic_img), dim=2
            )

        if x.roughness_img is not None:
            gfx_material.roughness_map = Texture(
                np.ascontiguousarray(x.roughness_img), dim=2
            )

        gfx_material.side = "front"

    elif x.shader == "unlit":
        gfx_material = MeshPhongMaterial()
        if x.base_color is not None:
            base_color = (
                np.array(x.base_color[:3]) / 255
            )  # Convert to normalized values
            gfx_material.color = Color(*base_color)

        gfx_material.shininess = getattr(x, "base_reflectance", 0.5)
        gfx_material.specular = Color(1, 1, 1)  # Default specular color

        if x.albedo_img is not None:
            gfx_material.map = Texture(np.ascontiguousarray(x.albedo_img), dim=2)

        gfx_material.side = "front"

    else:
        raise NotImplementedError(f"Shader type {x.shader} is not supported.")

    return gfx_material



"""Utils for the Buffer and Texture classes."""

import time
import sys
import logging
from math import ceil, log2

import numpy as np

logger = logging.getLogger("pygfx")


def get_element_format_from_numpy_array(array):
    """Get the per-element format specifier from a numpy array.
    Returns None if the format appears to be a structured array.
    Raises an error if GPU-incompatible dtypes are used (64 bit).
    """

    # Uniform buffers are scalars with a structured dtype.
    # But can also create storage buffers with complex formats.
    if array.dtype.kind not in "iuf":
        return None

    # GPUs generally don't support 64-bit buffers or textures.
    # Note: the Python docs say that l and L are 32 bit, but converting
    # a int64 numpy array to a memoryview gives a format of 'l' instead
    # of 'q' on some systems/configs? So we need to check the itemsize.
    if array.itemsize == 8:
        raise ValueError(
            f"A dtype of {array.dtype.name} is not supported for buffers, use a 32-bit variant instead."
        )

    return array.dtype.str.lstrip("<>=|")


SYS_ENDIANNESS = "<" if sys.byteorder == "little" else ">"


def is_little_endian(arr):
    """Get whether the given array is little endian."""
    byteorder = arr.dtype.byteorder
    if byteorder == "=":
        byteorder = SYS_ENDIANNESS
    elif byteorder == "|":  # single-byte dtype
        byteorder = "<"
    return byteorder == "<"


def make_little_endian(arr):
    """Get a copy of the array that has the same dtype but little endian."""
    return arr.astype(arr.dtype.newbyteorder("<"))


def check_data_is_clean_for_performance(kind, arr):
    """Check that data is c_contiguous and little endian. Raise an error if not."""
    missing_props = []
    if not is_little_endian(arr):
        missing_props.append("little endian")
    if not arr.flags.c_contiguous:
        missing_props.append("c_contiguous")
    if missing_props:
        raise ValueError(
            f"Given {kind} data is not {', '.join(missing_props)} (enforced because force_contiguous is set)."
        )


def get_alignment_multiplier(bytes_per_element=1, align=16):
    bytes_per_element = bytes_per_element or 1
    factor_max = log2(align)
    if not factor_max.is_integer():
        raise ValueError("align must be factor of two")
    for factor in range(int(factor_max) + 1):
        mult = 2**factor
        extra = (mult * bytes_per_element) % align
        if not extra:
            return mult


def calculate_buffer_chunk_size(buffer_size, **kwargs):
    """Same as calculate_texture_chunk_size, but for just one dimension."""
    res = calculate_texture_chunk_size((int(buffer_size), 1, 1), **kwargs)
    return res[0]


def calculate_texture_chunk_size(
    tex_size,
    *,
    bytes_per_element=1,
    byte_align=16,
    target_chunk_count=32,
    min_chunk_bytes=2**8,
    max_chunk_bytes=2**20,
):
    """Calculate the sizes to divide a texture in chunks.

    This algorithm tries to approach the target_chunk_count, divide the chunks equally
    over the dimensions, take alignment into account, as well as bounds on the chunk size.
    """
    # Check / normalize inputs
    tex_size = tuple(int(x) for x in tex_size)
    assert len(tex_size) == 3
    bytes_per_element = int(bytes_per_element)
    byte_align = int(byte_align)
    target_chunk_count = int(target_chunk_count)

    # Calculate the element-alignment to apply for the first dimension
    element_align = get_alignment_multiplier(bytes_per_element, byte_align)

    min_chunk_size_x = int(min_chunk_bytes / bytes_per_element)
    min_chunk_sizes = [
        min_chunk_size_x,
        max(1, int(min_chunk_size_x / tex_size[0])),
        max(1, int(min_chunk_size_x / tex_size[0] / tex_size[1])),
    ]

    max_chunk_elements = max_chunk_bytes / bytes_per_element

    # Prepare result
    chunk_size = [1, 1, 1]

    for index in (0, 1, 2):
        min_chunk_size = min_chunk_sizes[index]
        max_chunk_size = tex_size[index]

        # Get approximate chunk size, with limits applied
        approx_chunk_size = max_chunk_size / target_chunk_count
        approx_chunk_size = min(approx_chunk_size, max_chunk_elements)
        approx_chunk_size = max(approx_chunk_size, min_chunk_size)

        # Get the approximate chunk count, and snap the approximate
        # chunk_size to it. This is to have a more or less equal
        # distribution of the chunks (i.e. avoiding a half-chunk at the
        # end). Note that we don't try to get an exact integer number of
        # chunks, becaus this_might be hard/impossible (max_chunk_size might
        # be a prime number). A ceil works better here, otherwise a chunk size
        # can easily snap to the full size for smaller dimensions.
        approx_chunk_count = ceil(max_chunk_size / approx_chunk_size)
        approx_chunk_size = max_chunk_size / approx_chunk_count

        if index == 0:
            # Special care for x
            # If the texture is 3D, taking non-full rows (in x) is doubly-non-contiguous.
            # Benchmarks have shown that its better to not chunk in x in this case.
            if tex_size[2] > 1:
                this_chunk_size = tex_size[index]
            else:
                this_chunk_size = (
                    ceil(approx_chunk_size / element_align) * element_align
                )
        else:
            this_chunk_size = ceil(approx_chunk_size)

        # Apply final limits
        this_chunk_size = min(max_chunk_size, max(1, this_chunk_size))

        # Save
        chunk_size[index] = this_chunk_size
        max_chunk_elements /= this_chunk_size

    return tuple(chunk_size)


class ChunkBlock:
    """Little helper object."""

    __slots__ = ["nx", "ny", "nz", "x", "y", "z"]

    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z
        self.nx = 1
        self.ny = 1
        self.nz = 1

    def __repr__(self):
        return f"<Chunkblock ({self.x}, {self.y}, {self.z}), ({self.nx}, {self.ny}, {self.nz})>"

    def get_offset(self):
        return self.x, self.y, self.z

    def get_size(self):
        return self.nx, self.ny, self.nz


def get_merged_blocks_from_mask_1d(chunk_mask):
    """Algorithm to get a list of chunk descriptions from an 1D mask, with chunks merged.

    Return list of (offset, size) tuples.
    """

    blocks = []
    size = chunk_mask.size
    i = 0
    while i < size:
        if chunk_mask[i]:
            x = i
            nx = 1
            i += 1
            while i < size and chunk_mask[i]:
                nx += 1
                i += 1
            blocks.append((x, nx))
        else:
            i += 1
    return blocks


def get_merged_blocks_from_mask_3d(chunk_mask):
    """Algorithm to get a list of chunk descriptions from a 3D mask, with chunks merged.

    Returns a list of objects having fields x, y, z, nx, ny, nz.
    """

    # This algorithm only needs a 3D mask, making it easy to test and maintain.
    #
    # The problem can be defined relatively easy: given a 3D boolean mask, get a
    # list of blocks representing the elements that are "on". Neighbouring
    # on-elements must be combined in a single block (merging). It is ok if a
    # few off-elements are present in a block (aggresive merging). It is not ok
    # to omit any on-elements.
    #
    # From benchmarks we learned that textures suffer a larger performance
    # penalty for chunking than buffers do, except for chunks split in the last
    # dimension. This makes sense because it results in non-contiguous pieces of
    # data. Therefore the non-contiguous dimension is merged in full when it
    # contains sufficient on-values.
    #
    # This implementation leverages numpy to generate chunks, which I found
    # was faster (for average chunk counts) than a more Python-isch approach.

    t0 = time.perf_counter()  # noqa

    chunk_count_z, chunk_count_y, chunk_count_x = chunk_mask.shape

    # Get dimensionality, so we know whether to more agrressily merge chunks to get a contiguous chunk.
    ndims = 1
    if chunk_mask.shape[1] > 1:
        ndims = 2
    if chunk_mask.shape[0] > 1:
        ndims = 3

    # Note: don't fill gaps between chunks; this only hurts performance!
    # Even for the non-contiguous dimensions.

    # Get a mask to merge all chunks along a dim
    if ndims >= 2:
        mask_x = chunk_mask.sum(axis=2) >= 0.25 * chunk_count_x
    if ndims == 3:
        mask_y = chunk_mask.sum(axis=1) >= 0.25 * chunk_count_y

    chunks = []

    while True:
        # Get start of next block
        zz, yy, xx = np.where(chunk_mask)
        if not zz.size:
            break

        z, y, x = zz[0], yy[0], xx[0]
        nz, ny, nx = 1, 1, 1

        # Increase block size (i.e. merging)
        still_changing = True
        while still_changing:
            still_changing = False

            # Extend chunk in x-dimension
            while x + nx < chunk_count_x:
                extra = chunk_mask[z : z + nz, y : y + ny, x + nx : x + nx + 1]
                count = np.count_nonzero(extra)
                if count >= 0.5 * ny * nz:
                    nx += 1
                    still_changing = True
                else:
                    break

            # Extend chunk in y-dimension
            while y + ny < chunk_count_y:
                extra = chunk_mask[z : z + nz, y + ny : y + ny + 1, x : x + nx]
                count = np.count_nonzero(extra)
                if count >= 0.5 * nx * nz:
                    ny += 1
                    still_changing = True
                else:
                    break

            # Extend chunk in z-dimension
            while z + nz < chunk_count_z:
                extra = chunk_mask[z + nz : z + nz + 1, y : y + ny, x : x + nx]
                count = np.count_nonzero(extra)
                if count >= 0.5 * nx * ny:
                    nz += 1
                    still_changing = True
                else:
                    break

        # Extend to full size if its sufficiently filled
        if ndims >= 2:
            if mask_x[z, y] or nx >= 0.25 * chunk_count_x:
                x, nx = 0, chunk_count_x
        if ndims == 3:
            if mask_y[z, x] or ny >= 0.25 * chunk_count_y:
                y, ny = 0, chunk_count_y

        # Create the detected chunk
        chunk = ChunkBlock(x, y, z)
        chunk.nx = nx
        chunk.ny = ny
        chunk.nz = nz
        chunks.append(chunk)

        # Update the mask
        chunk_mask[z : z + nz, y : y + ny, x : x + nx] = False

    # print(chunk_mask.shape, len(chunks), f"{(time.perf_counter() - t0)*1000:0.2f}ms")
    return chunks



"""
The renderstate object defines details on how objects are rendered,
related to statefull stuff like lights in the scene and blender targets.
"""

import weakref
import numpy as np
import wgpu

from ....objects import (
    PointLight,
    DirectionalLight,
    SpotLight,
    AmbientLight,
)
from ....resources import Buffer
from ....utils import array_from_shadertype

from .pipeline import Binding
from .utils import generate_uniform_struct
from .shared import get_shared


# This cache is only used so renderstate objects can be re-used.
# The instances must be stored somewhere to prevent them from being deleted.
_renderstate_instance_cache = weakref.WeakValueDictionary()


def get_renderstate(light_dict, blender):
    """Get the renderstate object for the given scene and blender."""

    # Convert args to states
    light_state = (
        light_dict["point_lights"],
        light_dict["directional_lights"],
        light_dict["spot_lights"],
        light_dict["ambient_color"],
    )
    blend_state = (blender,)
    combined_state = light_state, blend_state

    # Create renderstate object and prepare it with the current state
    hash = CombinedRenderState.state_to_hash(*combined_state)
    ob = CombinedRenderState.obtain_from_hash(hash)
    ob.prepare_for_draw(*combined_state)

    return ob


# ----------


class BaseRenderState:
    """Object for internal use, that represents a certain state related to
    rendering, that affects the shader and/or pipeline.

    We distinguish the following three concepts: args -> state -> hash.

    The args is the incoming argument, e.g. the scene object. The state
    is derived from that, e.g. lists of light objects.
    """

    _hash = None

    @classmethod
    def obtain_from_hash(cls, hash):
        # Could do this in an __init__, but I like how this is more explicit
        ob = _renderstate_instance_cache.get(hash, None)
        if ob is None:
            ob = cls(*hash)
            ob._hash = hash
            _renderstate_instance_cache[hash] = ob
        return ob

    @property
    def hash(self):
        return self._hash

    # subclasses must implement

    @classmethod
    def state_to_hash(cls, *state):
        """Convert state-tuple to a hash-tuple."""
        raise NotImplementedError()

    def __init__(self, *hash):
        raise NotImplementedError()

    def prepare_for_draw(self, *state):
        raise NotImplementedError()

    def get_shader_kwargs(self, bind_group_index=1):
        raise NotImplementedError()


class LightRenderState(BaseRenderState):
    """Represents the rendering state for a specific number of lights (of each type)."""

    @classmethod
    def state_to_hash(
        cls, point_lights, directional_lights, spot_lights, ambient_color
    ):
        return (len(point_lights), len(directional_lights), len(spot_lights))

    def __init__(self, point_lights_count, dir_lights_count, spot_lights_count):
        self.device = get_shared().device

        # Store counts
        self.point_lights_count = point_lights_count
        self.dir_lights_count = dir_lights_count
        self.spot_lights_count = spot_lights_count

        # Note: we could make this configurable. Would probably have to be a global setting.
        self.shadow_map_size = (1024, 1024)

        # Create list of binding objects
        self.bindings = []
        self._setup_light_resources()

    def _setup_light_resources(self):
        # The ambient light binding is easy, because ambient lights can be combined on CPU

        self.ambient_lights_buffer = Buffer(
            array_from_shadertype(AmbientLight.uniform_type), force_contiguous=True
        )
        self.bindings.append(
            Binding(
                "u_ambient_light",
                "buffer/uniform",
                self.ambient_lights_buffer,
                structname="AmbientLight",
            )
        )

        # A bit more work for the directional lights

        self.directional_lights_buffer = None
        self.directional_lights_shadow_texture = None
        if self.dir_lights_count > 0:
            self.directional_lights_buffer = Buffer(
                array_from_shadertype(
                    DirectionalLight.uniform_type, self.dir_lights_count
                ),
                force_contiguous=True,
            )
            self.bindings.append(
                Binding(
                    "u_directional_lights",
                    "buffer/uniform",
                    self.directional_lights_buffer,
                    structname="DirectionalLight",
                )
            )
            self.directional_lights_shadow_texture = self.device.create_texture(
                size=(
                    self.shadow_map_size[0],
                    self.shadow_map_size[1],
                    self.dir_lights_count,
                ),
                usage=wgpu.TextureUsage.RENDER_ATTACHMENT
                | wgpu.TextureUsage.TEXTURE_BINDING,
                format="depth32float",
            )
            self.directional_lights_shadow_texture_views = [
                self.directional_lights_shadow_texture.create_view(base_array_layer=i)
                for i in range(self.dir_lights_count)
            ]
            self.bindings.append(
                Binding(
                    "u_shadow_map_dir_light",
                    "shadow_texture/2d-array",
                    self.directional_lights_shadow_texture.create_view(
                        dimension="2d-array"
                    ),
                )
            )

        # Similar logic for the point lights

        self.point_lights_buffer = None
        self.point_lights_shadow_texture = None
        if self.point_lights_count > 0:
            self.point_lights_buffer = Buffer(
                array_from_shadertype(PointLight.uniform_type, self.point_lights_count),
                force_contiguous=True,
            )
            self.bindings.append(
                Binding(
                    "u_point_lights",
                    "buffer/uniform",
                    self.point_lights_buffer,
                    structname="PointLight",
                )
            )
            self.point_lights_shadow_texture = self.device.create_texture(
                size=(
                    self.shadow_map_size[0],
                    self.shadow_map_size[1],
                    self.point_lights_count * 6,
                ),
                usage=wgpu.TextureUsage.RENDER_ATTACHMENT
                | wgpu.TextureUsage.TEXTURE_BINDING,
                format="depth32float",
            )
            self.point_lights_shadow_texture_views = [
                self.point_lights_shadow_texture.create_view(base_array_layer=i)
                for i in range(self.point_lights_count * 6)
            ]
            self.bindings.append(
                Binding(
                    "u_shadow_map_point_light",
                    "shadow_texture/cube-array",
                    self.point_lights_shadow_texture.create_view(
                        dimension="cube-array"
                    ),
                )
            )

        # And the spot lights

        self.spot_lights_buffer = None
        self.spot_lights_shadow_texture = None
        if self.spot_lights_count > 0:
            self.spot_lights_buffer = Buffer(
                array_from_shadertype(SpotLight.uniform_type, self.spot_lights_count),
                force_contiguous=True,
            )
            self.bindings.append(
                Binding(
                    "u_spot_lights",
                    "buffer/uniform",
                    self.spot_lights_buffer,
                    structname="SpotLight",
                )
            )
            self.spot_lights_shadow_texture = self.device.create_texture(
                size=(
                    self.shadow_map_size[0],
                    self.shadow_map_size[1],
                    self.spot_lights_count,
                ),
                usage=wgpu.TextureUsage.RENDER_ATTACHMENT
                | wgpu.TextureUsage.TEXTURE_BINDING,
                format="depth32float",
            )
            self.spot_lights_shadow_texture_views = [
                self.spot_lights_shadow_texture.create_view(base_array_layer=i)
                for i in range(self.spot_lights_count)
            ]
            self.bindings.append(
                Binding(
                    "u_shadow_map_spot_light",
                    "shadow_texture/2d-array",
                    self.spot_lights_shadow_texture.create_view(dimension="2d-array"),
                )
            )

        # We only need a sampler if we have shadow-casting lights

        if self.dir_lights_count + self.point_lights_count + self.spot_lights_count > 0:
            self.shadow_sampler = self.device.create_sampler(
                mag_filter=wgpu.FilterMode.linear,
                min_filter=wgpu.FilterMode.linear,
                compare=wgpu.CompareFunction.less_equal,
            )
            self.bindings.append(
                Binding(
                    "u_shadow_sampler",
                    "shadow_sampler/comparison",
                    self.shadow_sampler,
                )
            )

    def get_shader_kwargs(self, bind_group_index=1):
        """Get shader template kwargs specific to the renderstate.
        Used by the pipeline to complete the shader.
        """
        light_definitions = self._get_light_structs_code()
        light_definitions += self._get_light_vars_code(bind_group_index)
        kwargs = {
            "num_dir_lights": self.dir_lights_count,
            "num_spot_lights": self.spot_lights_count,
            "num_point_lights": self.point_lights_count,
            "light_definitions": light_definitions,
        }
        return kwargs

    def _get_light_structs_code(self):
        """Generate the wgsl that defines the light structs."""
        light_struct = []
        for binding in self.bindings:
            if binding.type.startswith("buffer/"):
                struct_code = generate_uniform_struct(
                    binding.resource.data.dtype, binding.structname
                )
                light_struct.append(struct_code)
        return "\n".join(light_struct)

    def _get_light_vars_code(self, bind_group_index):
        """Generate the wgsl that defines the uniforms, textures and sampler bindings."""
        codes = []
        for slot, binding in enumerate(self.bindings):
            if binding.type.startswith("buffer/"):
                uniform_type = (
                    f"array<{binding.structname}, {binding.resource.data.shape[0]}>"  # array of struct
                    if isinstance(binding.resource, Buffer)
                    and binding.resource.data.shape  # Buffer.items > 1
                    else binding.structname
                )
                code = f"""
                @group({bind_group_index}) @binding({slot})
                var<uniform> {binding.name}: {uniform_type};
                """.rstrip()
                codes.append(code)
            elif binding.type.startswith("shadow_texture/"):
                dim = binding.type.split("/")[-1].replace("-", "_")
                code = f"""
                @group({bind_group_index}) @binding({slot})
                var {binding.name}: texture_depth_{dim};
                """.rstrip()
                codes.append(code)
            elif binding.type.startswith("shadow_sampler/"):
                code = f"""
                @group({bind_group_index}) @binding({slot})
                var {binding.name}: sampler_comparison;
                """.rstrip()
                codes.append(code)

        return "\n".join(codes)

    def prepare_for_draw(
        self, point_lights, directional_lights, spot_lights, ambient_color
    ):
        """Update the contents of the uniform buffers for the lights,
        and create texture views if needed.
        """

        # Make light info available to other parts of the pygfx engine
        self.lights = {
            "point_lights": point_lights,
            "directional_lights": directional_lights,
            "spot_lights": spot_lights,
            "ambient_color": ambient_color,
        }

        # Update ambient buffer

        ambient_lights_buffer = self.ambient_lights_buffer
        if not np.all(ambient_lights_buffer.data["color"][:3] == ambient_color):
            ambient_lights_buffer.data["color"].flat = ambient_color
            ambient_lights_buffer.update_range(0, 1)

        # We update the uniform buffers of the lights below. These buffers
        # are not actually used directly but copied to the renderstate's buffer.
        # Seems like a detour, but kindof the simplest solution still.

        # Update directional light buffers

        if self.dir_lights_count > 0:
            dir_lights_buffer = self.directional_lights_buffer
            for i, light in enumerate(directional_lights):
                light._gfx_update_uniform_buffer()
                if light.cast_shadow:
                    light.shadow._gfx_update_uniform_buffer(light)
                    light.shadow._wgpu_tex_view = (
                        self.directional_lights_shadow_texture_views[i]
                    )
                if dir_lights_buffer.data[i] != light.uniform_buffer.data:
                    dir_lights_buffer.data[i] = light.uniform_buffer.data
                    dir_lights_buffer.update_range(i, 1)

        # Update point light buffers
        if self.point_lights_count > 0:
            point_lights_buffer = self.point_lights_buffer
            for i, light in enumerate(point_lights):
                light._gfx_update_uniform_buffer()
                if light.cast_shadow:
                    light.shadow._gfx_update_uniform_buffer(light)
                    light.shadow._wgpu_tex_view = [
                        self.point_lights_shadow_texture_views[i * 6 + face]
                        for face in range(6)
                    ]
                if point_lights_buffer.data[i] != light.uniform_buffer.data:
                    point_lights_buffer.data[i] = light.uniform_buffer.data
                    point_lights_buffer.update_range(i, 1)

        if self.spot_lights_count > 0:
            spot_lights_buffer = self.spot_lights_buffer
            for i, light in enumerate(spot_lights):
                light._gfx_update_uniform_buffer()
                if light.cast_shadow:
                    light.shadow._gfx_update_uniform_buffer(light)
                    light.shadow._wgpu_tex_view = self.spot_lights_shadow_texture_views[
                        i
                    ]
                if spot_lights_buffer.data[i] != light.uniform_buffer.data:
                    spot_lights_buffer.data[i] = light.uniform_buffer.data
                    spot_lights_buffer.update_range(i, 1)


class BlendRenderState(BaseRenderState):
    """Represents the rendering state for a specific blend mode."""

    @classmethod
    def state_to_hash(cls, blender):
        return (blender.hash,)

    def __init__(self, blend_hash):
        self.bindings = []

    def prepare_for_draw(self, blender):
        # Make blender available for other parts or the pygfx engine
        self.blender = blender

    def get_shader_kwargs(self, bind_group_index=1):
        return {}


class CombinedRenderState(BaseRenderState):
    """Represents the combined rendering state."""

    @classmethod
    def state_to_hash(cls, light_state, blend_state):
        light_hash = LightRenderState.state_to_hash(*light_state)
        blend_hash = BlendRenderState.state_to_hash(*blend_state)

        return light_hash, blend_hash

    def __init__(self, light_hash, blend_hash):
        self.device = get_shared().device

        self.light_renderstate = LightRenderState.obtain_from_hash(light_hash)
        self.blend_renderstate = BlendRenderState.obtain_from_hash(blend_hash)

        self.bindings = []
        self.bindings += self.light_renderstate.bindings
        self.bindings += self.blend_renderstate.bindings

        self.wgpu_bind_group_layout = None
        self.wgpu_bind_group = None
        if self.bindings:
            self.wgpu_bind_group_layout, self.wgpu_bind_group = self._collect_bindings()

    def prepare_for_draw(self, light_state, blend_state):
        # Delegate
        self.light_renderstate.prepare_for_draw(*light_state)
        self.blend_renderstate.prepare_for_draw(*blend_state)

        # Make state available for other parts or the pygfx engine
        self.lights = self.light_renderstate.lights
        self.blender = self.blend_renderstate.blender

    def get_shader_kwargs(self, bind_group_index=1):
        kwargs = {}
        kwargs.update(self.light_renderstate.get_shader_kwargs(bind_group_index))
        kwargs.update(self.blend_renderstate.get_shader_kwargs(bind_group_index))
        return kwargs

    # Stuff specific for the combined renderstate

    def _collect_bindings(self):
        """Group the bindings into a wgpu bind group."""
        bg_descriptor = []
        bg_layout_descriptor = []
        device = self.device

        for index, binding in enumerate(self.bindings):
            if binding.type.startswith("buffer/"):
                binding.resource._wgpu_usage |= wgpu.BufferUsage.UNIFORM
            binding_des, binding_layout_des = binding.get_bind_group_descriptors(index)
            bg_descriptor.append(binding_des)
            bg_layout_descriptor.append(binding_layout_des)

        bind_group_layout = device.create_bind_group_layout(
            entries=bg_layout_descriptor
        )
        bind_group = device.create_bind_group(
            layout=bind_group_layout, entries=bg_descriptor
        )
        return bind_group_layout, bind_group



from ..resources import Texture, TextureMap
from ._base import Material
from ..utils import assert_type


class ImageBasicMaterial(Material):
    """Rasterized image material.

    Parameters
    ----------
    clim : tuple
        The contrast limits to scale the data values with. Default (0, 1).
    map : TextureMap | Texture
        The texture map to turn the image values into its final color. Optional.
    gamma : float
        The gamma correction to apply to the image data. Default 1.
    interpolation : str
        The method to interpolate the image data. Either 'nearest' or 'linear'. Default 'nearest'.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class
        <pygfx.Material>`.

    """

    uniform_type = dict(
        Material.uniform_type,
        clim="2xf4",
        gamma="f4",
    )

    def __init__(
        self,
        clim=None,
        map=None,
        gamma=1.0,
        interpolation="nearest",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.map = map
        self.clim = clim
        self.gamma = gamma
        self.interpolation = interpolation

    @property
    def map(self):
        """The texture map to turn the image values into its final color.
        If None, the values themselves represents the color. The
        dimensionality of the texture map can be 1D, 2D or 3D, but
        should match the number of channels in the image.

        Note: for scientific data, it is usually to set wrap mode to 'clamp'.
        """
        return self._store.map

    @map.setter
    def map(self, map):
        assert_type("map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map, wrap="clamp")
        self._store.map = map

    @property
    def clim(self):
        """The contrast limits to scale the data values with. Default (0, 1)."""
        v1, v2 = self.uniform_buffer.data["clim"]
        return float(v1), float(v2)

    @clim.setter
    def clim(self, clim):
        # Check and store given clim
        if clim is None:
            clim = 0, 1
        clim = float(clim[0]), float(clim[1])
        # Update uniform data
        self.uniform_buffer.data["clim"] = clim
        self.uniform_buffer.update_full()

    @property
    def gamma(self):
        """The gamma correction to apply to the image data. Default 1."""
        return self.uniform_buffer.data["gamma"]

    @gamma.setter
    def gamma(self, value):
        value = float(value)
        if value <= 0:
            raise ValueError("gamma must be greater than 0")
        self.uniform_buffer.data["gamma"] = float(value)
        self.uniform_buffer.update_full()

    @property
    def interpolation(self):
        """The method to interpolate the image data. Either 'nearest' or 'linear'."""
        return self._store.interpolation

    @interpolation.setter
    def interpolation(self, value):
        assert value in ("nearest", "linear")
        self._store.interpolation = value



"""
Experimental support for compute shaders.
This should eventually be a more generic class, and move to wgpu.utils, or a new library.
See https://github.com/pygfx/wgpu-py/issues/704
"""

import time

from typing import Optional, Union
import wgpu
import pygfx as gfx


# TODO: move this into wgpu/pygfx/new lib
# TODO: ability to concatenate multiple steps
# TODO: add support for uniforms
class ComputeShader:
    """Abstraction for a compute shader.

    Parameters
    ----------
    wgsl : str
        The compute shader's code as WGSL.
    entry_point : str | None
        The name of the wgsl function that must be called.
        If the wgsl code has only one entry-point (a function marked with ``@compute``)
        this argument can be omitted.
    label : str | None
        The label for this shader. Used to set labels of underlying wgpu objects,
        and in debugging messages. If not set, use the entry_point.
    report_time : bool
        When set to True, will print the spent time to run the shader.
    """

    def __init__(
        self,
        wgsl,
        *,
        entry_point: Optional[str] = None,
        label: Optional[str] = None,
        report_time: bool = False,
    ):
        # Fixed
        self._wgsl = wgsl
        self._entry_point = entry_point
        self._label = label or entry_point or ""
        self._report_time = report_time

        # Things that can be changed via the API
        self._resources = {}
        self._constants = {}

        # Flag to keep track whether this object changed.
        # Note that this says nothing about the contents of buffers/textures used as input.
        self._changed = True

        # Internal variables
        self._device = None
        self._shader_module = None
        self._pipeline = None
        self._bind_group = None

    @property
    def changed(self) -> bool:
        """Whether the shader has been changed.

        This can be a new value for a constant, or a different resource.
        Note that this says nothing about the values inside a buffer or texture resource.
        This value is reset when ``dispatch()`` is called.
        """
        return self._changed

    def set_resource(
        self,
        index: int,
        resource: Union[gfx.Buffer, gfx.Texture, wgpu.GPUBuffer, wgpu.GPUTexture],
        *,
        clear=False,
    ):
        """Set a resource.

        Parameters
        ----------
        index : int
            The binding index to connect this resource to. (The group is hardcoded to zero for now.)
        resource : buffer | texture
            The buffer or texture to attach. Can be a wgpu or pygfx resource.
        clear : bool
            When set to True (only possible for a buffer), the resource is cleared to zeros
            right before running the shader.
        """
        # Check
        if not isinstance(index, int):
            raise TypeError(f"ComputeShader resource index must be int, not {index!r}.")
        if not isinstance(
            resource, (gfx.Buffer, gfx.Texture, wgpu.GPUBuffer, wgpu.GPUTexture)
        ):
            raise TypeError(
                f"ComputeShader resource value must be gfx.Buffer, gfx.Texture, wgpu.GPUBuffer, or wgpu.GPUTexture, not {resource!r}"
            )
        clear = bool(clear)
        if clear and not isinstance(
            resource, (gfx.Buffer, gfx.Texture, wgpu.GPUBuffer)
        ):
            raise ValueError("Can only clear a buffer, not a texture.")

        # Value to store
        new_value = resource, bool(clear)

        # Update if different
        old_value = self._resources.get(index)
        if new_value != old_value:
            if resource is None:
                self._resources.pop(index, None)
            else:
                self._resources[index] = new_value
            self._bind_group = None
            self._changed = True

    def set_constant(self, name: str, value: Union[bool, int, float, None]):
        """Set override constant.

        Setting override constants don't require shader recompilation, but does
        require re-creating the pipeline object. So it's less suited for things
        that change on every draw.
        """
        # NOTE: we could also provide support for uniform variables.
        # The override constants are nice and simple, but require the pipeline
        # to be re-created whenever a contant changes.

        # Check
        if not isinstance(name, str):
            raise TypeError(f"ComputeShader constant name must be str, not {name!r}.")
        if not (value is None or isinstance(value, (bool, int, float))):
            raise TypeError(
                f"ComputeShader constant value must be bool, int, float, or None, not {value!r}."
            )

        # Update if different
        old_value = self._constants.get(name)
        if value != old_value:
            if value is None:
                self._constants.pop(name, None)
            else:
                self._constants[name] = value
            self._pipeline = None
            self._changed = True

    def _get_native_resource(self, resource):
        if isinstance(resource, gfx.Resource):
            return gfx.renderers.wgpu.engine.update.ensure_wgpu_object(resource)
        return resource

    def _get_bindings_from_resources(self):
        bindings = []
        for index, (resource, _) in self._resources.items():
            # Get wgpu.GPUBuffer or wgpu.GPUTexture
            wgpu_object = self._get_native_resource(resource)
            if isinstance(wgpu_object, wgpu.GPUBuffer):
                bindings.append(
                    {
                        "binding": index,
                        "resource": {
                            "buffer": wgpu_object,
                            "offset": 0,
                            "size": wgpu_object.size,
                        },
                    }
                )
            elif isinstance(wgpu_object, wgpu.GPUTexture):
                bindings.append(
                    {
                        "binding": index,
                        "resource": wgpu_object.create_view(
                            usage=wgpu.TextureUsage.STORAGE_BINDING
                        ),
                    }
                )
            else:
                raise RuntimeError(f"Unexpected resource: {resource}")
        return bindings

    def dispatch(self, nx, ny=1, nz=1):
        """Dispatch the workgroups, i.e. run the shader."""
        nx, ny, nz = int(nx), int(ny), int(nz)

        # Reset
        self._changed = False

        # Get device
        if self._device is None:
            self._shader_module = None
            self._device = gfx.renderers.wgpu.Shared.get_instance().device
        device = self._device

        # Compile the shader
        if self._shader_module is None:
            self._pipeline = None
            self._shader_module = device.create_shader_module(
                label=self._label, code=self._wgsl
            )

        # Get the pipeline object
        if self._pipeline is None:
            self._bind_group = None
            self._pipeline = device.create_compute_pipeline(
                label=self._label,
                layout="auto",
                compute={
                    "module": self._shader_module,
                    "entry_point": self._entry_point,
                    "constants": self._constants,
                },
            )

        # Get the bind group object
        if self._bind_group is None:
            bind_group_layout = self._pipeline.get_bind_group_layout(0)
            bindings = self._get_bindings_from_resources()
            self._bind_group = device.create_bind_group(
                label=self._label, layout=bind_group_layout, entries=bindings
            )

        # Make sure that all used resources have a wgpu-representation, and are synced
        for resource, _ in self._resources.values():
            if isinstance(resource, gfx.Resource):
                gfx.renderers.wgpu.engine.update.update_resource(resource)

        t0 = time.perf_counter()

        # Start!
        command_encoder = device.create_command_encoder(label=self._label)

        # Maybe clear some buffers
        for resource, clear in self._resources.values():
            if clear:
                command_encoder.clear_buffer(self._get_native_resource(resource))

        # Do the compute pass
        compute_pass = command_encoder.begin_compute_pass()
        compute_pass.set_pipeline(self._pipeline)
        compute_pass.set_bind_group(0, self._bind_group)
        compute_pass.dispatch_workgroups(nx, ny, nz)
        compute_pass.end()

        # Submit!
        device.queue.submit([command_encoder.finish()])

        # Timeit
        if self._report_time:
            device._poll_wait()  # wait for the GPU to finish
            t1 = time.perf_counter()
            what = f"Computing {self._label!r}" if self._label else "Computing"
            print(f"{what} took {(t1 - t0) * 1000:0.1f} ms")



import wgpu  # only for flags/enums

from ....objects import Background
from ....materials import BackgroundMaterial, BackgroundImageMaterial
from ....resources import TextureMap

from .. import (
    register_wgpu_render_function,
    BaseShader,
    Binding,
    to_texture_format,
    GfxSampler,
    GfxTextureView,
    load_wgsl,
)


@register_wgpu_render_function(Background, BackgroundMaterial)
class BackgroundShader(BaseShader):
    type = "render"

    def get_bindings(self, wobject, shared):
        material = wobject.material

        bindings = {}

        # Uniforms
        bindings[0] = Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer)
        bindings[1] = Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer)
        bindings[2] = Binding("u_material", "buffer/uniform", material.uniform_buffer)

        if isinstance(material, BackgroundImageMaterial) and material.map is not None:
            if not isinstance(material.map, TextureMap):
                raise TypeError("material.map must be a TextureMap.")
            sampler = GfxSampler("linear", "repeat")
            # Select texture dimension
            texture = material.map.texture
            if texture.size[2] == 1:
                tex_view = GfxTextureView(texture, view_dim="2d")
                self["texture_dim"] = "2d"
            elif texture.size[2] == 6:
                tex_view = GfxTextureView(texture, view_dim="cube")
                self["texture_dim"] = "cube"
            else:
                raise ValueError(
                    "BackgroundImageMaterial.map size must be NxMx1 or NxMx6."
                )
            bindings[3] = Binding("r_sampler", "sampler/filtering", sampler, "FRAGMENT")
            bindings[4] = Binding("r_tex", "texture/auto", tex_view, "FRAGMENT")
            # Channels
            fmt = to_texture_format(texture.format)
            self["texture_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))
        else:
            self["texture_dim"] = ""

        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_strip,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        return {
            "indices": (4, 1),
        }

    def get_code(self):
        return load_wgsl("background.wgsl")



"""Helpers for visual debugging of scenes.

.. currentmodule:: pygfx.helpers

This module contains a collection of WorldObjects that can be useful
when debugging a scene or to create reference points within a scene.

.. autosummary::
    :toctree: helpers/
    :template: ../_templates/custom_layout.rst

    AxesHelper
    GridHelper
    BoxHelper
    TransformGizmo
    PointLightHelper
    DirectionalLightHelper
    SpotLightHelper
    Stats
    SkeletonHelper

"""

# ruff: noqa: F401

from ._axes import AxesHelper
from ._grid import GridHelper
from ._box import BoxHelper
from ._gizmo import TransformGizmo
from ._lights import (
    PointLightHelper,
    DirectionalLightHelper,
    SpotLightHelper,
)
from ._stats import Stats
from ._skeleton import SkeletonHelper



from math import pi, tan

import numpy as np
import pylinalg as la

from ..objects._base import WorldObject
from ..utils.transform import cached
from ._base import Camera


# Some general remarks on how the camera works:
#
# The width and height have a special meaning as they represent the size of the scene.
# Internally, we use the concept of 'extent', which is the mean of the width and height.
# The orbit camera uses the extent to determine the point around which to rotate, and the
# fly camera uses the extent to determine the movement speed.
#
# So although, with a fov > 0, the width and height do not affect the projection matrix,
# they still play a role. The controllers actually sets both the distance and the width/height.
# That way the user can change the fov at any time and transition smoothly.
#
# The orthograpic camera constraints the vof to zero. With a fov of zero it is not necessary to move the
# camera away from the scene in order to see it. This avoids depth issues for very long scenes like
# time-series plots. And avoids the scene clipping away (becoming invisible) of you zoom out too much.


class PerspectiveCamera(Camera):
    """A generic 3D camera with a configurable field of view (fov).

    Parameters
    ----------
    fov: float
        The field of view as an angle in degrees. Higher values give a
        wide-angle lens effect. This value is limited between 0 and
        179. If zero, it operates in orthographic mode. Default 50.
    aspect: float
        The desired aspect ratio (width divided by height), which determines the
        vision pyramid's boundaries. Default 1.
    width: float
        The width of the scene to view. If omitted or None, the width
        is derived from aspect and height.
    height: float
        The height of the scene to view. If omitted or None, the height
        is derived from aspect and width.
    zoom: float
        An additional zoom factor, equivalent to attaching a zoom lens.
    maintain_aspect: bool
        Whether the aspect ration is maintained as the window size changes.
        Default True. If false, the dimensions are stretched to fit the window.
    depth: float | None
        The reference size of the scene in the depth dimension. By default this value gets set
        to the average of ``width`` and ``height``. This happens on initialization and by ``show_pos()``, ``show_object()`` and ``show_rect()``.
        It is used to calculate a sensible depth range when ``depth_range`` is not set.
    depth_range: 2-tuple | None
        The explicit values for the near and far clip planes. If not given
        or None (the default), the clip planes ware calculated using ``fov`` and ``depth``.

    Notes
    -----

    When the fov is zero, the width and height determine the view of the camera.
    When the fov is nonzero, the width and height only specify the aspect ratio.
    However, the width and height still represent the size of the scene, and the controllers
    uses it to e.g. determine the orbit-point, or the speed of motion.

    When you use ``show_pos()``, ``show_object()``, or ``show_rect()``, there is no
    need to set the width or height.

    When you don't use the show methods, and the fov is zero, you need to set the width and height, and optionally the depth.
    """

    _fov_range = 0, 179

    def __init__(
        self,
        fov=50,
        aspect=1,
        *,
        width=None,
        height=None,
        zoom=1,
        maintain_aspect=True,
        depth=None,
        depth_range=None,
    ):
        super().__init__()

        self.fov = fov

        # Set width and height. Note that if both width and height are given, it overrides aspect
        aspect = aspect or 1
        if width is None and height is None:
            height = 100  # 100 produces safer near plane for most use-cases than 1
            width = height * aspect
        elif width is None:
            width = height * aspect
        elif height is None:
            height = width / aspect
        self.width = width
        self.height = height

        self.zoom = zoom
        self.maintain_aspect = maintain_aspect
        self.depth = depth or float(0.5 * (width + height))
        self.depth_range = depth_range

        # depth_range_auto =

        self.set_view_size(1, 1)

    @property
    def fov(self):
        """The field of view (in degrees), between 0-179."""
        return self._fov

    @fov.setter
    def fov(self, value):
        fov = float(value)
        fov = min(max(fov, self._fov_range[0]), self._fov_range[1])
        fov = fov_limit(fov)  # Constraint to numerically stable values
        self._fov = fov
        self.flag_update()

    @property
    def width(self):
        """The (minimum) width of the view-cube.
        Changing the width changes the aspect, but not the height.
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = float(value)
        self.flag_update()

    @property
    def height(self):
        """The (minimum) height of the view-cube.
        Changing the height changes the aspect, but not the width.
        """
        return self._height

    @height.setter
    def height(self, value):
        self._height = float(value)
        self.flag_update()

    @property
    def aspect(self):
        """The aspect ratio (width divided by height).

        The aspect determines the vision pyramid's boundaries. It can help fit
        the scene better to the window, and keep it fitting as the window is
        resized.

        Setting the aspect updates width and height such that their mean is
        unchanged.

        Note that ``show_object(match_aspect=True)`` also sets the aspect.
        """
        return self._width / self._height

    @aspect.setter
    def aspect(self, value):
        aspect = float(value)
        if aspect <= 0:
            raise ValueError("aspect must be > 0")
        extent = 0.5 * (self._width + self._height)
        self._height = 2 * extent / (1 + aspect)
        self._width = self._height * aspect
        self.flag_update()

    def _set_extent(self, extent):
        """Set the mean of width and height while maintaining aspect."""
        extent = float(extent)
        if extent <= 0:
            raise ValueError("extent must be > 0")
        aspect = self.aspect
        self._height = 2 * extent / (1 + aspect)
        self._width = self._height * aspect
        self.flag_update()

    @property
    def zoom(self):
        """The camera zoom level."""
        return self._zoom

    @zoom.setter
    def zoom(self, value):
        self._zoom = float(value)
        self.flag_update()

    @property
    def maintain_aspect(self):
        """Whether the aspect ration is maintained as the window size changes.

        Default True. Note that it only make sense to set this to False in
        combination with a panzoom controller.
        """
        return self._maintain_aspect

    @maintain_aspect.setter
    def maintain_aspect(self, value):
        self._maintain_aspect = bool(value)
        self.flag_update()

    @property
    def depth(self):
        """The reference size for the scene in the depth dimension.

        This is only used if ``depth_range`` is not set.

        This value is set by ``show_pos()``, ``show_object()`` and ``show_rect()``.
        It can be set by the user, but this is only recommended if you don't use any of the show methods.
        """
        return self._depth

    @depth.setter
    def depth(self, depth):
        depth = float(depth)
        if depth <= 0:
            raise ValueError("Camera.depth must both be > 0")
        self._depth = depth

    @property
    def depth_range(self):
        """The user-defined values for the near and far clip planes.

        By default this is None, causing the near and far planes to be calculated from ``fov`` and ``depth``.
        This is a convenience that works well for simple cases. The automatic depth range calculation assumes
        that the scene is confined, containing one or more objects to be inspected. The depth range is set quite
        a bit larger than the scene, so the scene does not disappear as you zoom out, and e.g. grids have a proper horizon.

        Cases where explicitly setting ``depth_range`` makes sense:

        * When you know your scene well, and just want to be explicit.
        * When the scene represents a larger world where there are near and distant objects.
        * When the automatically calculated depth range is suboptimal.
        * When the depth of your scene is much smaller/larger than (the average of the) width and height,
          e.g. in plotting. Although you can also set ``depth`` in this case.
        """
        return self._depth_range

    @depth_range.setter
    def depth_range(self, value):
        if value is None:
            self._depth_range = None
        elif isinstance(value, (tuple, list)) and len(value) == 2:
            self._depth_range = float(value[0]), float(value[1])
        else:
            raise TypeError("depth_range must be None or a 2-tuple.")
        self.flag_update()

    def _get_near_and_far_plane(self):
        # Dept range explicitly given?
        if self._depth_range:
            return self._depth_range

        depth = self._depth

        if self.fov > 0:
            # Scale near plane with the fov to compensate for the fact that with very small
            # fov you're probably looking at something in the far distance.
            f = fov_distance_factor(self.fov)
            # We want to be gentle with the factor for the near plane; making that value small will cost a lot of bits in the depth buffer.
            # The value for the far buffer affects the precision near the camerea much less.
            return depth * f / 100, depth * f * 10000
        else:
            # Look behind and in front in equal distance.
            # With a fov of 0, the depth precision is divided equally over the whole range. So being able to look
            # far in the distance, is *much* more costly than it is for perpective projection.
            # With a factor 100, you can zoom out until the scene is just a few pixes before it disappears.
            return (-100 * depth, +100 * depth)

    @property
    def near(self) -> float:
        """The location of the near clip plane.
        Use `depth_range` so overload the computed value, if necessary.
        """
        near, _ = self._get_near_and_far_plane()
        return near

    @property
    def far(self) -> float:
        """The location of the far clip plane.
        Use `depth_range` so overload the computed value, if necessary.
        """
        _, far = self._get_near_and_far_plane()
        return far

    def get_state(self):
        """Get the state of the camera as a dict.

        The fields contain "position", "rotation", "scale", and
        "reference_up", representing the camera's transform. The scale
        is typically not used, but included for completeness. Further,
        the following properties are included: "fov", "width", "height",
        "zoom", "maintain_aspect", and "depth_range".

        """
        return {
            "position": self.local.position.copy(),
            "rotation": self.local.rotation.copy(),
            "scale": self.local.scale.copy(),
            "reference_up": self.world.reference_up.copy(),
            "fov": self.fov,
            "width": self.width,
            "height": self.height,
            "depth": self.depth,
            "zoom": self.zoom,
            "maintain_aspect": self.maintain_aspect,
            "depth_range": self.depth_range,
        }

    def set_state(self, state):
        """Set the state of the camera from a dict.

        Accepted fields are the same as in ``get_state()``. In addition,
        the fields ``x``, ``y``, and ``z`` are also accepted to set the
        position along a singular dimension.

        """
        # Set the more complex props
        for key, value in state.items():
            if key == "position":
                self.local.position = value
            if key in ("x", "y", "z"):
                setattr(self.local, key, value)
            elif key == "scale":
                self.local.scale = value
            elif key == "rotation":
                self.local.rotation = value
            elif key == "reference_up":
                self.world.reference_up = value
            elif key in (
                "fov",
                "width",
                "height",
                "depth",
                "zoom",
                "maintain_aspect",
                "depth_range",
            ):
                # Simple props
                setattr(self, key, value)
        self.flag_update()

    def _update_projection_matrix(self):
        zoom_factor = self._zoom
        near, far = self._get_near_and_far_plane()

        view_aspect = self._view_size[0] / self._view_size[1]

        if self._view_offset is not None:
            # The view_offset should override the aspect, via its full (virtual) size
            view_aspect = (
                self._view_offset["full_width"] / self._view_offset["full_height"]
            )

        if self.fov > 0:
            # Get the reference width / height
            size = 2 * near * tan(pi / 180 * 0.5 * self.fov) / zoom_factor
            # Pre-apply the reference aspect ratio
            height = 2 * size / (1 + self.aspect)
            width = height * self.aspect
            # Increase either the width or height, depending on the view size
            if self._maintain_aspect:
                if self.aspect < view_aspect:
                    width *= view_aspect / self.aspect
                else:
                    height *= self.aspect / view_aspect
            # Calculate bounds
            top = +0.5 * height
            bottom = -0.5 * height
            left = -0.5 * width
            right = +0.5 * width
            # Set matrices
            projection_matrix = la.mat_perspective(
                left, right, top, bottom, near, far, depth_range=(0, 1)
            )

        else:
            # The reference view plane is scaled with the zoom factor
            width = self.width / zoom_factor
            height = self.height / zoom_factor
            # Increase either the width or height, depending on the viewport shape
            aspect = width / height
            if self._maintain_aspect:
                if aspect < view_aspect:
                    width *= view_aspect / aspect
                else:
                    height *= aspect / view_aspect
            # Calculate bounds
            bottom = -0.5 * height
            top = +0.5 * height
            left = -0.5 * width
            right = +0.5 * width
            # Set matrices
            projection_matrix = la.mat_orthographic(
                left, right, top, bottom, near, far, depth_range=(0, 1)
            )

        projection_matrix.flags.writeable = False
        return projection_matrix

    def show_pos(self, target, *, up=None, depth=None):
        """Look at the given position or object.

        This is similar to `look_at()` but it also sets the width and height
        (while honoring aspect). The camera's position is not changed, but it's
        'reference point' does. E.g. with an orbit controller, it orbits around
        the given target.

        Parameters
        ----------
        target: WorldObject or (x, y, z)
            The target to point the camera towards.
        up: 3-tuple
            If given, set ``camera.world.reference_up`` to the given value.
        depth: float
            The reference size for the scene in the depth dimension. If not
            given, the ``camera.depth`` is set based on ``.width`` and ``.height``.
        """

        # Get pos from target
        if isinstance(target, WorldObject):
            pos = target.local.position
        else:
            pos = np.asarray(target)

        if pos.shape != (3,):
            raise ValueError("Expected position to have 3 values.")

        # Look at the provided position, taking up into account
        if up is not None:
            self.world.reference_up = up
        self.look_at(pos)

        # Update extent
        distance = la.vec_dist(pos, self.local.position)
        f = fov_distance_factor(self.fov)
        if f == 0:
            extent = distance
        else:
            extent = distance / f
        self._set_extent(extent)

        # Lock the reference extent for consistent deph range
        self.depth = extent if depth is None else depth

    def show_object(
        self,
        target: WorldObject,
        view_dir=None,
        *,
        up=None,
        scale=1,
        match_aspect=False,
        depth=None,
    ):
        """Position and orientate the camera such that the given target in is in view.

        Sets the position and rotation of the camera, and adjusts
        its width and height to the target's size (while honoring aspect).

        The fit is such that as the object is rotated (or as the camera rotates around the object)
        the object still fits inside the viewport, which is usually good for 3D scenes.

        When setting ``match_aspect``, the bounding box of the object is matched
        to the screen, and the camera's ``aspect`` adjusted accordingly. This
        gives a tight fit that is usually preferred for 2D scenes (with the
        panzoom-controller).

        Parameters
        ----------
        target: WorldObject or sphere (x, y, z, r)
            The object to look at.
        view_dir: 3-tuple of float
            Look at the object from this direction. If not given or None,
            uses the current view direction.
        up: 3-tuple
            If given, set ``camera.world.reference_up`` to the given value.
        scale: float
            Scale the size of what's shown. Default 1.
        match_aspect: bool
            Whether to match the camera's ``width`` and ``height`` to the
            target's bounding box so it tighly fits the viewport. Useful with a
            PanZoomController, less so with a 3D scene.
        depth: float
            The reference size for the scene in the depth dimension. If not
            given, the ``camera.depth`` is set based on ``.width`` and ``.height``.
        """

        if up is None:
            up = self.world.reference_up
        else:
            up = np.asarray(up)
            self.world.reference_up = up

        # Get bounding sphere from target
        bbox = None
        if isinstance(target, WorldObject):
            bsphere = target.get_world_bounding_sphere()
            bbox = target.get_world_bounding_box()
            if bsphere is None:
                raise ValueError(
                    "Given target does not have a bounding sphere, you should probably just provide a sphere (x, y, z, r) yourself."
                )
        elif isinstance(target, (tuple, list, np.ndarray)) and len(target) == 4:
            bsphere = tuple(target)
        else:
            raise TypeError(
                "show_object target must be a WorldObject, or a (x, y, z, radius) tuple."
            )

        # Obtain view direction
        if view_dir is None:
            rotation = self.local.rotation
            view_dir = la.vec_transform_quat((0, 0, -1), rotation)
        elif isinstance(view_dir, (tuple, list, np.ndarray)) and len(view_dir) == 3:
            view_dir = tuple(view_dir)
        else:
            raise TypeError(f"Expected view_dir to be sequence, not {view_dir}")
        view_dir = la.vec_normalize(view_dir)

        # Do the math ...
        view_pos = bsphere[:3]
        radius = max(0.0, bsphere[3]) or 1.0
        extent = radius * 2 * scale

        # Apply
        distance = fov_distance_factor(self.fov) * extent
        self.world.position = view_pos - view_dir * distance
        self.world.forward = view_dir
        self._set_extent(extent)

        if match_aspect and bbox is not None:
            # Re-calculate width and height using the aligned bbox, so that the
            # contents keep fitting snugly as the viewport is resized.
            bbox = la.aabb_transform(bbox, self.world.inverse_matrix)
            extent_xy = (bbox[1, :2] - bbox[0, :2]) * scale
            self.width = float(extent_xy[0])
            self.height = float(extent_xy[1])
            # Adust distance to match the new extent (the direction is unchanged)
            extent = 0.5 * (self._width + self._height)
            distance = fov_distance_factor(self.fov) * extent
            self.world.position = view_pos - view_dir * distance

        # Note that when fov == 0, distance == 0, and thus self.world.position == view_pos

        # Lock the reference extent for consistent deph range
        self.depth = extent if depth is None else depth

    def show_rect(
        self, left, right, top, bottom, *, view_dir=None, up=None, depth=None
    ):
        """Position the camera such that the given rectangle is in view.

        The rectangle represents a plane in world coordinates, centered
        at the origin of the world, and rotated to be orthogonal to the
        view_dir.

        Sets the position and rotation of the camera, and adjusts
        width and height to the rectangle (thus also changing the aspect).

        This method is mainly intended for viewing 2D data, especially
        when `maintain_aspect` is set to False, and is convenient
        for setting the initial view before attaching a PanZoomController.

        See ``show_object(..., match_aspect=True)`` for a more automatic approach.

        Parameters
        ----------
        left: float
            The left boundary of the plane to show.
        right: float
            The right boundary of the plane to show.
        top: float
             The top boundary of the plane to show.
        bottom: float
             The bottom boundary of the plane to show.
        view_dir: 3-tuple of float
            Look at the rectangle from this direction. If not given or None,
            uses the current view direction.
        up: 3-tuple
            If given, set ``camera.world.reference_up`` to the given value.
        depth: float
            The reference size for the scene in the depth dimension. If not
            given, the ``camera.depth`` is set based on ``.width`` and ``.height``.
        """

        if up is not None:
            self.world.reference_up = up

        # Obtain view direction
        if view_dir is None:
            view_dir = la.vec_transform_quat((0, 0, -1), self.world.rotation)
        else:
            view_dir = la.vec_normalize(view_dir)

        # Set bounds, note that this implicitly sets width, height (and aspect)
        self.width = right - left
        self.height = bottom - top
        extent = 0.5 * (self.width + self.height)
        # First move so we view towards the origin with the correct vector
        # Note that when fov == 0, distance == 0, and thus self.world.position == (0, 0, 0)
        distance = fov_distance_factor(self.fov) * extent
        self.world.position = (0, 0, 0) - view_dir * distance
        self.world.forward = view_dir

        # Now we have a rotation that we can use to orient our rect
        rotation = self.world.rotation
        position = self.world.position
        offset = 0.5 * (left + right), 0.5 * (top + bottom), 0
        new_position = position + la.vec_transform_quat(offset, rotation)
        self.world.position = new_position

        # Lock the reference extent for consistent deph range
        self.depth = extent if depth is None else depth

    @cached
    def frustum(self):
        """Corner positions of the viewing frustum in world space.

        Returns
        -------
        frustum : ndarray, [2, 4, 3]
            The coordinates of the frustum. The first axis corresponds to the
            frustum's plane (near, far), the second to the corner within the
            plane ((left, bottom), (right, bottom), (right, top), (left, top)),
            and the third to the world position of that corner.

        """
        ndc_corners = np.array([(-1, -1), (1, -1), (1, 1), (-1, 1)])
        depths = np.array((0, 1))[:, None]
        local_corners = la.vec_unproject(
            ndc_corners, self.projection_matrix, depth=depths
        )
        world_corners = la.vec_transform(local_corners, self.world.matrix)
        world_corners.flags.writeable = False
        return world_corners


def fov_limit(fov):
    # Don't allow values between 0 and 1, as it becomes numerically unstable.
    # Same for values beyond 179.
    # Zero is allowed, meaning orthographic projection
    if fov < 0.5:
        fov = 0.0
    elif fov < 1.0:
        fov = 1.0
    elif fov > 179:
        fov = 179
    return fov


def fov_distance_factor(fov):
    # It's important that controller and camera use the same distance calculations
    if fov > 0:
        fov_rad = fov * pi / 180
        factor = 0.5 / tan(0.5 * fov_rad)
        factor = max(factor, 0.0)  #  prevent negative numbers
    else:
        factor = 0.0
    return factor
